#!/usr/bin/env python
"""
_Fileset_t_

Unit tests for the WMBS Fileset class.
"""

import threading
import unittest

from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMQuality.TestInit import TestInit


class FilesetTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Setup the database and logging connection.  Try to create all of the
        WMBS tables.
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)

        myThread = threading.currentThread()
        self.daofactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        locationAction = self.daofactory(classname="Locations.New")
        locationAction.execute(siteName="site1", pnn="T2_CH_CERN")
        return

    def tearDown(self):
        """
        _tearDown_

        Drop all the WMBS tables.
        """
        self.testInit.clearDatabase()

    def testCreateDeleteExists(self):
        """
        _testCreateDeleteExists_

        Create a delete a fileset object while also using the exists() method
        to determine the the create() and delete() methods succeeded.
        """
        testFileset = Fileset(name="TestFileset")

        assert testFileset.exists() is False, \
            "ERROR: Fileset exists before it was created"

        testFileset.create()

        assert testFileset.exists() >= 0, \
            "ERROR: Fileset does not exist after it was created"

        testFileset.delete()

        assert testFileset.exists() is False, \
            "ERROR: Fileset exists after it was deleted"

        return

    def testCreateDeleteLongChar(self):
        """
        _testCreateDeleteLongChar_

        Create and delete a fileset name with 750 characters, InnoDB tables
        are limited to 767 chars
        """
        # fileset name with 1000 chars
        fsetName = "test_Fileset_with_50_character_and_anything_else_" * 20
        testFileset = Fileset(name=fsetName)

        self.assertFalse(testFileset.exists())
        testFileset.create()
        self.assertTrue(testFileset.exists() >= 0)
        testFileset.delete()
        self.assertFalse(testFileset.exists())

        # fileset name with 1500 chars
        fsetName = "test_Fileset_with_50_character_and_anything_else_" * 30
        testFileset = Fileset(name=fsetName)

        self.assertFalse(testFileset.exists())
        with self.assertRaises(Exception):
            testFileset.create()

        return

    def testCreateTransaction(self):
        """
        _testCreateTransaction_

        Create a Fileset and commit it to the database and then roll back the
        transaction.  Use the fileset's exists() method to verify that it
        doesn't exist in the database before create() is called, that is does
        exist after create() is called and that it does not exist after the
        transaction is rolled back.
        """
        myThread = threading.currentThread()
        myThread.transaction.begin()

        testFileset = Fileset(name="TestFileset")

        assert testFileset.exists() is False, \
            "ERROR: Fileset exists before it was created"

        testFileset.create()

        assert testFileset.exists() >= 0, \
            "ERROR: Fileset does not exist after it was created"

        myThread.transaction.rollback()

        assert testFileset.exists() is False, \
            "ERROR: Fileset exists after transaction was rolled back."

        return

    def testDeleteTransaction(self):
        """
        _testDeleteTransaction_

        Create a fileset and commit it to the database.  Delete the fileset
        and verify that it is no longer in the database using the exists()
        method.  Rollback the transaction and verify with the exists() method
        that the fileset is in the database.
        """
        testFileset = Fileset(name="TestFileset")

        assert testFileset.exists() is False, \
            "ERROR: Fileset exists before it was created"

        testFileset.create()

        assert testFileset.exists() >= 0, \
            "ERROR: Fileset does not exist after it was created"

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testFileset.delete()

        assert testFileset.exists() is False, \
            "ERROR: Fileset exists after it was deleted"

        myThread.transaction.rollback()

        assert testFileset.exists() >= 0, \
            "ERROR: Fileset doesn't exist after transaction was rolled back."

        return

    def testLoad(self):
        """
        _testLoad_

        Test retrieving fileset metadata via the id and the
        name.
        """
        testFilesetA = Fileset(name="TestFileset")
        testFilesetA.create()

        testFilesetB = Fileset(name=testFilesetA.name)
        testFilesetB.load()
        testFilesetC = Fileset(id=testFilesetA.id)
        testFilesetC.load()

        self.assertTrue(isinstance(testFilesetB.id, int), "Fileset id is not an int.")
        self.assertTrue(isinstance(testFilesetC.id, int), "Fileset id is not an int.")

        assert testFilesetB.id == testFilesetA.id, \
            "ERROR: Load from name didn't load id"

        assert testFilesetC.name == testFilesetA.name, \
            "ERROR: Load from id didn't load name"

        # Verify we can add files to a loaded fileset.
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileA.create()
        testFilesetC.addFile(testFileA)
        testFilesetC.commit()

        testFilesetC.loadData()
        assert len(testFilesetC.files) == 1, \
            "Error: Wrong number of files."

        testFilesetA.delete()
        return

    def testLoadData(self):
        """
        _testLoadData_

        Test saving and loading all fileset information.
        """
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileA.addRun(Run(1, *[45]))
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileB.addRun(Run(1, *[45]))
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileC.addRun(Run(1, *[45]))
        testFileA.create()
        testFileB.create()
        testFileC.create()

        testFilesetA = Fileset(name="TestFileset")
        testFilesetA.create()

        testFilesetA.addFile(testFileA)
        testFilesetA.addFile(testFileB)
        testFilesetA.addFile(testFileC)
        testFilesetA.commit()

        testFilesetB = Fileset(name=testFilesetA.name)
        testFilesetB.loadData()
        testFilesetC = Fileset(id=testFilesetA.id)
        testFilesetC.loadData()

        assert testFilesetB.id == testFilesetA.id, \
            "ERROR: Load from name didn't load id"

        assert testFilesetC.name == testFilesetA.name, \
            "ERROR: Load from id didn't load name"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetB.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetC.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        testFilesetA.delete()
        testFileA.delete()
        testFileB.delete()
        testFileC.delete()

    def testGetFiles(self):
        """
        _testGetFiles_

        Create a fileset with three files and exercise the getFiles() method to
        make sure that all the results it returns are consistant.
        """
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileA.addRun(Run(1, *[45]))
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileB.addRun(Run(1, *[45]))
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileC.addRun(Run(1, *[45]))
        testFileA.create()
        testFileB.create()
        testFileC.create()

        testFilesetA = Fileset(name="TestFileset")
        testFilesetA.create()

        testFilesetA.addFile(testFileA)
        testFilesetA.addFile(testFileB)
        testFilesetA.addFile(testFileC)
        testFilesetA.commit()

        filesetFiles = testFilesetA.getFiles(type="list")

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in filesetFiles:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Not all files in fileset"

        filesetFiles = testFilesetA.getFiles(type="set")

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in filesetFiles:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Not all files in fileset"

        filesetLFNs = testFilesetA.getFiles(type="lfn")

        goldenLFNs = [testFileA["lfn"], testFileB["lfn"], testFileC["lfn"]]
        for filesetLFN in filesetLFNs:
            assert filesetLFN in goldenLFNs, \
                "ERROR: Unknown lfn in fileset"
            goldenLFNs.remove(filesetLFN)

        assert len(goldenLFNs) == 0, \
            "ERROR: Not all lfns in fileset"

        filesetIDs = testFilesetA.getFiles(type="id")

        goldenIDs = [testFileA["id"], testFileB["id"], testFileC["id"]]
        for filesetID in filesetIDs:
            assert filesetID in goldenIDs, \
                "ERROR: Unknown id in fileset"
            goldenIDs.remove(filesetID)

        assert len(goldenIDs) == 0, \
            "ERROR: Not all ids in fileset"

    def testFileCreate(self):
        """
        _testFileCreate_

        Create several files and add them to the fileset.  Test to make sure
        that the commit() fileset method will add the files to the database
        if they are not in the database.  Also verify that files are correctly
        marked as available.
        """
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileA.addRun(Run(1, *[45]))
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileB.addRun(Run(1, *[45]))
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileC.addRun(Run(1, *[45]))
        testFileC.create()

        testWorkflowA = Workflow(spec="spec1.xml", owner="Hassen",
                                 name="wf001", task="sometask")
        testWorkflowA.create()

        testFilesetA = Fileset(name="TestFileset")
        testFilesetA.create()

        testSubscriptionA = Subscription(fileset=testFilesetA,
                                         workflow=testWorkflowA)
        testSubscriptionA.create()

        testFilesetA.addFile(testFileA)
        testFilesetA.addFile(testFileB)
        testFilesetA.addFile(testFileC)
        testFilesetA.commit()

        testFilesetB = Fileset(name=testFilesetA.name)
        testFilesetB.loadData()
        testFilesetC = Fileset(id=testFilesetA.id)
        testFilesetC.loadData()

        assert testFilesetB.id == testFilesetA.id, \
            "ERROR: Load from name didn't load id"

        assert testFilesetC.name == testFilesetA.name, \
            "ERROR: Load from id didn't load name"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetB.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testSubscriptionA.filesOfStatus("Available"):
            assert filesetFile["lfn"] in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile["lfn"])

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetC.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        return

    def testFileCreateTransaction(self):
        """
        _testFileCreateTransaction_

        Create several files and add them to a fileset.  Commit the fileset
        and the files to the database, verifying that they can loaded back
        from the database.  Rollback the transaction to the point after the
        fileset has been created buy before the files have been associated with
        the filset.  Load the filesets from the database again and verify that
        they do not have any files.
        """
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileA.addRun(Run(1, *[45]))
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileB.addRun(Run(1, *[45]))
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileC.addRun(Run(1, *[45]))
        testFileC.create()

        testFilesetA = Fileset(name="TestFileset")
        testFilesetA.create()

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testFilesetA.addFile(testFileA)
        testFilesetA.addFile(testFileB)
        testFilesetA.addFile(testFileC)
        testFilesetA.commit()

        testFilesetB = Fileset(name=testFilesetA.name)
        testFilesetB.loadData()
        testFilesetC = Fileset(id=testFilesetA.id)
        testFilesetC.loadData()

        assert testFilesetB.id == testFilesetA.id, \
            "ERROR: Load from name didn't load id"

        assert testFilesetC.name == testFilesetA.name, \
            "ERROR: Load from id didn't load name"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetB.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetC.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        myThread.transaction.rollback()

        testFilesetB.loadData()
        testFilesetC.loadData()

        assert len(testFilesetB.files) == 0, \
            "ERROR: Fileset B has too many files"

        assert len(testFilesetC.files) == 0, \
            "ERROR: Fileset C has too many files"

        testFilesetA.delete()
        testFileA.delete()
        testFileB.delete()
        testFileC.delete()

    def testMarkOpen(self):
        """
        _testMarkOpen_

        Test that setting the openess of a fileset in the constructor works as
        well as changing it with the markOpen() method.
        """
        testFilesetA = Fileset(name="TestFileset1", is_open=False)
        testFilesetA.create()
        testFilesetB = Fileset(name="TestFileset2", is_open=True)
        testFilesetB.create()

        testFilesetC = Fileset(name=testFilesetA.name)
        testFilesetC.load()
        testFilesetD = Fileset(name=testFilesetB.name)
        testFilesetD.load()

        assert testFilesetC.open is False, \
            "ERROR: FilesetC should be closed."

        assert testFilesetD.open is True, \
            "ERROR: FilesetD should be open."

        testFilesetA.markOpen(True)
        testFilesetB.markOpen(False)

        testFilesetE = Fileset(name=testFilesetA.name)
        testFilesetE.load()
        testFilesetF = Fileset(name=testFilesetB.name)
        testFilesetF.load()

        assert testFilesetE.open is True, \
            "ERROR: FilesetE should be open."

        assert testFilesetF.open is False, \
            "ERROR: FilesetF should be closed."

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        openFilesetDAO = daoFactory(classname="Fileset.ListOpen")
        openFilesetNames = openFilesetDAO.execute()

        assert len(openFilesetNames) == 1, \
            "ERROR: Too many open filesets."

        assert "TestFileset1" in openFilesetNames, \
            "ERROR: Wrong fileset listed as open."

        return

    def testFilesetClosing(self):
        """
        _testFilesetClosing_

        Verify the proper operation of the closable fileset DAO object.  A
        fileset is closable if:
          - All of the subscriptions that feed it has completed processing all
            files in their input fileset
          _ All of the jobs for feeder subscriptions have completed
          - The fileset that feeds the subscription is closed
          - The workflow for the subscription is fully injected.
        """
        testOutputFileset1 = Fileset(name="TestOutputFileset1")
        testOutputFileset1.create()
        testOutputFileset2 = Fileset(name="TestOutputFileset2")
        testOutputFileset2.create()
        testOutputFileset3 = Fileset(name="TestOutputFileset3")
        testOutputFileset3.create()
        testOutputFileset4 = Fileset(name="TestOutputFileset4")
        testOutputFileset4.create()

        testMergedOutputFileset1 = Fileset(name="TestMergedOutputFileset1")
        testMergedOutputFileset1.create()
        testMergedOutputFileset2 = Fileset(name="TestMergedOutputFileset2")
        testMergedOutputFileset2.create()
        testMergedOutputFileset3 = Fileset(name="TestMergedOutputFileset3")
        testMergedOutputFileset3.create()
        testMergedOutputFileset4 = Fileset(name="TestMergedOutputFileset4")
        testMergedOutputFileset4.create()

        testFilesetOpen = Fileset(name="TestFilesetOpen", is_open=True)
        testFilesetOpen.create()
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetOpen.addFile(testFileA)
        testFilesetOpen.addFile(testFileB)
        testFilesetOpen.commit()

        testFilesetClosed = Fileset(name="TestFilesetClosed", is_open=False)
        testFilesetClosed.create()
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileD = File(lfn="/this/is/a/lfnD", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetClosed.addFile(testFileC)
        testFilesetClosed.addFile(testFileD)
        testFilesetClosed.commit()

        testWorkflow1 = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf001", task="sometask")
        testWorkflow1.create()
        testWorkflow1.addOutput("out1", testOutputFileset1, testMergedOutputFileset1)
        testWorkflow1.addOutput("out2", testOutputFileset2, testMergedOutputFileset2)

        testWorkflow2 = Workflow(spec="spec2.xml", owner="Steve",
                                 name="wf002", task="sometask")
        testWorkflow2.create()
        testWorkflow2.addOutput("out3", testOutputFileset3, testMergedOutputFileset3)

        testWorkflow3 = Workflow(spec="spec4.xml", owner="Steve",
                                 name="wf004", task="sometask")
        testWorkflow3.create()
        testWorkflow3.addOutput("out4", testOutputFileset4, testMergedOutputFileset4)

        testSubscription1 = Subscription(fileset=testFilesetClosed,
                                         workflow=testWorkflow1)
        testSubscription1.create()
        testSubscription1.completeFiles([testFileC, testFileD])
        testSubscription2 = Subscription(fileset=testFilesetOpen,
                                         workflow=testWorkflow2)
        testSubscription2.create()
        testSubscription2.completeFiles([testFileA, testFileB])
        testSubscription3 = Subscription(fileset=testFilesetClosed,
                                         workflow=testWorkflow3)
        testSubscription3.create()

        testJobGroup = JobGroup(subscription=testSubscription1)
        testJobGroup.create()

        testJob = Job(name="TestJob1")
        testJob.create(testJobGroup)
        testJob["state"] = "executing"

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)

        changeStateDAO = daoFactory(classname="Jobs.ChangeState")
        changeStateDAO.execute(jobs=[testJob])

        closableFilesetDAO = daoFactory(classname="Fileset.ListClosable")
        closableFilesets = closableFilesetDAO.execute()

        assert len(closableFilesets) == 0, \
            "Error: There should be no closable filesets."

        testJob["state"] = "cleanout"
        changeStateDAO.execute(jobs=[testJob])
        closableFilesets = closableFilesetDAO.execute()

        assert len(closableFilesets) == 0, \
            "Error: There should be no closable filesets."

        injected = daoFactory(classname="Workflow.MarkInjectedWorkflows")
        injected.execute(names=["wf001", "wf002", "wf003"], injected=True)

        closableFilesets = closableFilesetDAO.execute()
        goldenFilesets = ["TestOutputFileset1", "TestOutputFileset2"]

        for closableFileset in closableFilesets:
            newFileset = Fileset(id=closableFileset)
            newFileset.load()

            assert newFileset.name in goldenFilesets, \
                "Error: Unknown closable fileset"

            goldenFilesets.remove(newFileset.name)

        assert len(goldenFilesets) == 0, \
            "Error: Filesets are missing"
        return

    def testFilesetClosing2(self):
        """
        _testFilesetClosing2_

        Verify that fileset closing works correctly in the case where multiple
        subscriptions feed into a single fileset.
        """
        testOutputFileset1 = Fileset(name="TestOutputFileset1")
        testOutputFileset1.create()
        testOutputFileset2 = Fileset(name="TestOutputFileset2")
        testOutputFileset2.create()

        testMergedOutputFileset1 = Fileset(name="TestMergedOutputFileset1")
        testMergedOutputFileset1.create()
        testMergedOutputFileset2 = Fileset(name="TestMergedOutputFileset2")
        testMergedOutputFileset2.create()

        testFilesetOpen = Fileset(name="TestFilesetOpen", is_open=True)
        testFilesetOpen.create()
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetOpen.addFile(testFileA)
        testFilesetOpen.addFile(testFileB)
        testFilesetOpen.commit()

        testFilesetClosed1 = Fileset(name="TestFilesetClosed1", is_open=False)
        testFilesetClosed1.create()
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileD = File(lfn="/this/is/a/lfnD", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetClosed1.addFile(testFileC)
        testFilesetClosed1.addFile(testFileD)
        testFilesetClosed1.commit()

        testFilesetClosed2 = Fileset(name="TestFilesetClosed2", is_open=False)
        testFilesetClosed2.create()
        testFileE = File(lfn="/this/is/a/lfnE", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileF = File(lfn="/this/is/a/lfnF", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetClosed2.addFile(testFileE)
        testFilesetClosed2.addFile(testFileF)
        testFilesetClosed2.commit()

        testFilesetClosed3 = Fileset(name="TestFilesetClosed3", is_open=False)
        testFilesetClosed3.create()
        testFileG = File(lfn="/this/is/a/lfnG", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileH = File(lfn="/this/is/a/lfnH", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetClosed3.addFile(testFileG)
        testFilesetClosed3.addFile(testFileH)
        testFilesetClosed3.commit()

        testWorkflow1 = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf001", task="sometask")
        testWorkflow1.create()
        testWorkflow1.addOutput("out1", testOutputFileset1, testMergedOutputFileset1)

        testWorkflow2 = Workflow(spec="spec2.xml", owner="Steve",
                                 name="wf002", task="sometask")
        testWorkflow2.create()
        testWorkflow2.addOutput("out2", testOutputFileset2, testMergedOutputFileset2)

        testSubscription1 = Subscription(fileset=testFilesetOpen,
                                         workflow=testWorkflow1)
        testSubscription1.create()
        testSubscription1.completeFiles([testFileA, testFileB])
        testSubscription2 = Subscription(fileset=testFilesetClosed1,
                                         workflow=testWorkflow1)
        testSubscription2.create()

        testSubscription3 = Subscription(fileset=testFilesetClosed2,
                                         workflow=testWorkflow2)
        testSubscription3.create()
        testSubscription3.completeFiles([testFileE, testFileF])
        testSubscription4 = Subscription(fileset=testFilesetClosed3,
                                         workflow=testWorkflow2)
        testSubscription4.create()
        testSubscription4.completeFiles([testFileG, testFileH])

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        injected = daoFactory(classname="Workflow.MarkInjectedWorkflows")
        injected.execute(names=["wf001", "wf002"], injected=True)
        closableFilesetDAO = daoFactory(classname="Fileset.ListClosable")
        closableFilesets = closableFilesetDAO.execute()

        goldenFilesets = ["TestOutputFileset2"]

        for closableFileset in closableFilesets:
            newFileset = Fileset(id=closableFileset)
            newFileset.load()

            assert newFileset.name in goldenFilesets, \
                "Error: Unknown closable fileset"

            goldenFilesets.remove(newFileset.name)

        assert len(goldenFilesets) == 0, \
            "Error: Filesets are missing"
        return

    def testFilesetClosing3(self):
        """
        _testFilesetClosing3_

        Verify that fileset closing works correctly in the case where multiple
        subscriptions feed into a single fileset and accounts for running jobs
        correctly.
        """
        testOutputFileset1 = Fileset(name="TestOutputFileset1")
        testOutputFileset1.create()
        testOutputFileset2 = Fileset(name="TestOutputFileset2")
        testOutputFileset2.create()

        testMergedOutputFileset1 = Fileset(name="TestMergedOutputFileset1")
        testMergedOutputFileset1.create()
        testMergedOutputFileset2 = Fileset(name="TestMergedOutputFileset2")
        testMergedOutputFileset2.create()

        testFilesetOpen = Fileset(name="TestFilesetOpen", is_open=False)
        testFilesetOpen.create()
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetOpen.addFile(testFileA)
        testFilesetOpen.addFile(testFileB)
        testFilesetOpen.commit()

        testFilesetClosed1 = Fileset(name="TestFilesetClosed1", is_open=False)
        testFilesetClosed1.create()
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileD = File(lfn="/this/is/a/lfnD", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetClosed1.addFile(testFileC)
        testFilesetClosed1.addFile(testFileD)
        testFilesetClosed1.commit()

        testFilesetClosed2 = Fileset(name="TestFilesetClosed2", is_open=False)
        testFilesetClosed2.create()
        testFileE = File(lfn="/this/is/a/lfnE", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileF = File(lfn="/this/is/a/lfnF", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetClosed2.addFile(testFileE)
        testFilesetClosed2.addFile(testFileF)
        testFilesetClosed2.commit()

        testFilesetClosed3 = Fileset(name="TestFilesetClosed3", is_open=False)
        testFilesetClosed3.create()
        testFileG = File(lfn="/this/is/a/lfnG", size=1024,
                         events=20, checksums={'cksum': 3})
        testFileH = File(lfn="/this/is/a/lfnH", size=1024,
                         events=20, checksums={'cksum': 3})
        testFilesetClosed3.addFile(testFileG)
        testFilesetClosed3.addFile(testFileH)
        testFilesetClosed3.commit()

        testWorkflow1 = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf001", task="sometask")
        testWorkflow1.create()
        testWorkflow1.addOutput("out1", testOutputFileset1, testMergedOutputFileset1)

        testWorkflow2 = Workflow(spec="spec2.xml", owner="Steve",
                                 name="wf002", task="sometask")
        testWorkflow2.create()
        testWorkflow2.addOutput("out2", testOutputFileset2, testMergedOutputFileset2)

        testSubscription1 = Subscription(fileset=testFilesetOpen,
                                         workflow=testWorkflow1)
        testSubscription1.create()
        testSubscription1.completeFiles([testFileA, testFileB])
        testSubscription2 = Subscription(fileset=testFilesetClosed1,
                                         workflow=testWorkflow1)
        testSubscription2.create()

        testJobGroup = JobGroup(subscription=testSubscription2)
        testJobGroup.create()
        testJob = Job(name="TestJob1")
        testJob.create(testJobGroup)

        testSubscription3 = Subscription(fileset=testFilesetClosed2,
                                         workflow=testWorkflow2)
        testSubscription3.create()
        testSubscription3.completeFiles([testFileE, testFileF])
        testSubscription4 = Subscription(fileset=testFilesetClosed3,
                                         workflow=testWorkflow2)
        testSubscription4.create()
        testSubscription4.completeFiles([testFileG, testFileH])

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        injected = daoFactory(classname="Workflow.MarkInjectedWorkflows")
        injected.execute(names=["wf001", "wf002"], injected=True)
        closableFilesetDAO = daoFactory(classname="Fileset.ListClosable")
        closableFilesets = closableFilesetDAO.execute()

        goldenFilesets = ["TestOutputFileset2"]

        for closableFileset in closableFilesets:
            newFileset = Fileset(id=closableFileset)
            newFileset.load()

            assert newFileset.name in goldenFilesets, \
                "Error: Unknown closable fileset"

            goldenFilesets.remove(newFileset.name)

        assert len(goldenFilesets) == 0, \
            "Error: Filesets are missing"
        return

    def testFilesetClosing4(self):
        """
        _testFilesetClosing4_

        Verify that fileset closing works correctly when a workflow completly
        fails out and does not produce any files.
        """
        testOutputFileset1 = Fileset(name="TestOutputFileset1")
        testOutputFileset1.create()
        testOutputFileset2 = Fileset(name="TestOutputFileset2")
        testOutputFileset2.create()
        testOutputFileset3 = Fileset(name="TestOutputFileset3")
        testOutputFileset3.create()

        testMergedOutputFileset1 = Fileset(name="TestMergedOutputFileset1")
        testMergedOutputFileset1.create()
        testMergedOutputFileset2 = Fileset(name="TestMergedOutputFileset2")
        testMergedOutputFileset2.create()
        testMergedOutputFileset3 = Fileset(name="TestMergedOutputFileset3")
        testMergedOutputFileset3.create()

        testOutputFileset1.markOpen(False)
        testOutputFileset2.markOpen(True)
        testOutputFileset3.markOpen(True)

        testInputFileset = Fileset(name="TestInputFileset")
        testInputFileset.create()
        testInputFileset.markOpen(False)

        testWorkflow1 = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf001", task="sometask")
        testWorkflow1.create()
        testWorkflow1.addOutput("out1", testOutputFileset1, testMergedOutputFileset1)

        testWorkflow2 = Workflow(spec="spec2.xml", owner="Steve",
                                 name="wf002", task="sometask")
        testWorkflow2.create()
        testWorkflow2.addOutput("out2", testOutputFileset2, testMergedOutputFileset2)

        testWorkflow3 = Workflow(spec="spec3.xml", owner="Steve",
                                 name="wf003", task="sometask")
        testWorkflow3.create()
        testWorkflow3.addOutput("out3", testOutputFileset3, testMergedOutputFileset3)

        testSubscription1 = Subscription(fileset=testInputFileset,
                                         workflow=testWorkflow1)
        testSubscription1.create()
        testSubscription2 = Subscription(fileset=testOutputFileset1,
                                         workflow=testWorkflow2)
        testSubscription2.create()
        testSubscription3 = Subscription(fileset=testOutputFileset2,
                                         workflow=testWorkflow3)
        testSubscription3.create()

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        injected = daoFactory(classname="Workflow.MarkInjectedWorkflows")
        injected.execute(names=["wf001", "wf002", "wf003"], injected=True)
        closableFilesetDAO = daoFactory(classname="Fileset.ListClosable")
        closableFilesets = closableFilesetDAO.execute()

        assert len(closableFilesets) == 1, \
            "Error: Wrong number of closable filesets"

        assert closableFilesets[0] == testOutputFileset2.id, \
            "Error: Wrong fileset is marked as closable."

        return

    def testFilesetClosing5(self):
        """
        _testFilesetClosing5_

        Verify that fileset closing works in the case where one cleanup
        subscription is used to cleanup files from all the other merge
        subscriptions in the request.
        """
        inputFileset = Fileset(name="InputFileset")
        inputFileset.create()
        inputFileset.markOpen(False)
        cleanupFileset = Fileset(name="CleanupFileset")
        cleanupFileset.create()
        cleanupFileset.markOpen(True)

        testOutputFileset1 = Fileset(name="TestOutputFileset1")
        testOutputFileset1.create()
        testOutputFileset1.markOpen(True)
        testOutputFileset2 = Fileset(name="TestOutputFileset2")
        testOutputFileset2.create()
        testOutputFileset2.markOpen(True)
        testOutputFileset3 = Fileset(name="TestOutputFileset3")
        testOutputFileset3.create()
        testOutputFileset3.markOpen(True)

        cleanupWorkflow = Workflow(spec="spec1.xml", owner="Steve",
                                   name="wf001", task="cleanup")
        cleanupWorkflow.create()

        testWorkflow1 = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf001", task="sometask1")
        testWorkflow1.create()
        testWorkflow1.addOutput("out1", testOutputFileset1)
        testWorkflow1.addOutput("out1", cleanupFileset)

        testWorkflow2 = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf001", task="sometask2")
        testWorkflow2.create()
        testWorkflow2.addOutput("out1", testOutputFileset2)
        testWorkflow2.addOutput("out1", cleanupFileset)

        testWorkflow3 = Workflow(spec="spec1.xml", owner="Steve",
                                 name="wf001", task="sometask3")
        testWorkflow3.create()
        testWorkflow3.addOutput("out1", testOutputFileset3)
        testWorkflow3.addOutput("out1", cleanupFileset)

        cleanupSubscription = Subscription(fileset=cleanupFileset,
                                           workflow=cleanupWorkflow)
        cleanupSubscription.create()

        testSubscription1 = Subscription(fileset=inputFileset,
                                         workflow=testWorkflow1)
        testSubscription1.create()
        testSubscription2 = Subscription(fileset=testOutputFileset1,
                                         workflow=testWorkflow2)
        testSubscription2.create()
        testSubscription3 = Subscription(fileset=testOutputFileset2,
                                         workflow=testWorkflow3)
        testSubscription3.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileA.addRun(Run(1, *[45]))
        testFileA.create()
        inputFileset.addFile(testFileA)
        inputFileset.commit()

        testJobGroupA = JobGroup(subscription=testSubscription1)
        testJobGroupA.create()

        testJobA = Job(name="TestJobA", files=[testFileA])
        testJobA.create(testJobGroupA)
        testJobA["state"] = "executing"

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        injected = daoFactory(classname="Workflow.MarkInjectedWorkflows")
        injected.execute(names=["wf001"], injected=True)
        changeStateDAO = daoFactory(classname="Jobs.ChangeState")
        changeStateDAO.execute(jobs=[testJobA])

        closableFilesetDAO = daoFactory(classname="Fileset.ListClosable")
        closableFilesets = closableFilesetDAO.execute()

        self.assertEqual(len(closableFilesets), 0,
                         "Error: There should be no closable filesets.")

        testSubscription1.completeFiles(testFileA)
        testJobA["state"] = "cleanout"
        changeStateDAO.execute(jobs=[testJobA])

        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileB.addRun(Run(1, *[45]))
        testFileB.create()
        testOutputFileset1.addFile(testFileB)
        testOutputFileset1.commit()
        cleanupFileset.addFile(testFileB)
        cleanupFileset.commit()

        closableFilesets = closableFilesetDAO.execute()
        self.assertEqual(len(closableFilesets), 1,
                         "Error: There should only be one closable fileset.")
        self.assertEqual(closableFilesets[0], testOutputFileset1.id,
                         "Error: Output fileset one should be closable.")

        testOutputFileset1.markOpen(False)

        testJobGroupB = JobGroup(subscription=testSubscription2)
        testJobGroupB.create()

        testJobB = Job(name="TestJobB", files=[testFileB])
        testJobB.create(testJobGroupB)
        testJobB["state"] = "cleanout"
        changeStateDAO.execute(jobs=[testJobB])

        testSubscription2.completeFiles([testFileB])
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileC.addRun(Run(1, *[45]))
        testFileC.create()
        testOutputFileset2.addFile(testFileC)
        testOutputFileset2.commit()
        cleanupFileset.addFile(testFileC)
        cleanupFileset.commit()

        closableFilesets = closableFilesetDAO.execute()
        self.assertEqual(len(closableFilesets), 1,
                         "Error: There should only be one closable fileset.")
        self.assertEqual(closableFilesets[0], testOutputFileset2.id,
                         "Error: Output fileset two should be closable.")

        testOutputFileset2.markOpen(False)

        testJobGroupC = JobGroup(subscription=testSubscription3)
        testJobGroupC.create()

        testJobC = Job(name="TestJobC", files=[testFileC])
        testJobC.create(testJobGroupC)
        testJobC["state"] = "cleanout"
        changeStateDAO.execute(jobs=[testJobC])

        testSubscription3.completeFiles([testFileC])
        testFileD = File(lfn="/this/is/a/lfnD", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileD.addRun(Run(1, *[45]))
        testFileD.create()
        testOutputFileset3.addFile(testFileD)
        testOutputFileset3.commit()
        cleanupFileset.addFile(testFileD)
        cleanupFileset.commit()

        closableFilesets = closableFilesetDAO.execute()
        self.assertEqual(len(closableFilesets), 2,
                         "Error: There should only be two closable filesets.")
        self.assertTrue(testOutputFileset3.id in closableFilesets,
                        "Error: Output fileset three should be closable.")
        self.assertTrue(cleanupFileset.id in closableFilesets,
                        "Error: Cleanup fileset should be closable.")
        return

    def testBulkAddDAO(self):
        """
        _testBulkAddDAO_

        Verify that the Fileset.BulkAdd DAO correct adds files to multiple
        filesets.
        """
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileA.addRun(Run(1, *[45]))
        testFileA.create()
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileB.addRun(Run(1, *[45]))
        testFileB.create()
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileC.addRun(Run(1, *[45]))
        testFileC.create()

        testFileD = File(lfn="/this/is/a/lfnD", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileD.addRun(Run(1, *[45]))
        testFileD.create()

        testWorkflowA = Workflow(spec="spec1.xml", owner="Hassen",
                                 name="wf001", task="sometask")
        testWorkflowA.create()

        testFilesetA = Fileset(name="TestFilesetA")
        testFilesetA.create()
        testFilesetB = Fileset(name="TestFilesetB")
        testFilesetB.create()

        testSubscriptionA = Subscription(fileset=testFilesetA,
                                         workflow=testWorkflowA)
        testSubscriptionA.create()
        testSubscriptionB = Subscription(fileset=testFilesetB,
                                         workflow=testWorkflowA)
        testSubscriptionB.create()

        myThread = threading.currentThread()
        myThread.transaction.begin()

        binds = [{"fileid": testFileA["id"], "fileset": testFilesetA.id},
                 {"fileid": testFileB["id"], "fileset": testFilesetA.id},
                 {"fileid": testFileC["id"], "fileset": testFilesetA.id},
                 {"fileid": testFileB["id"], "fileset": testFilesetB.id},
                 {"fileid": testFileC["id"], "fileset": testFilesetB.id},
                 {"fileid": testFileD["id"], "fileset": testFilesetB.id}]

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        bulkAddAction = daoFactory(classname="Fileset.BulkAdd")
        bulkAddAction.execute(binds=binds)

        testFilesetA.loadData()
        testFilesetB.loadData()

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetA.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testSubscriptionA.filesOfStatus("Available"):
            assert filesetFile["lfn"] in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile["lfn"])

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileB, testFileC, testFileD]
        for filesetFile in testFilesetB.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileB, testFileC, testFileD]
        for filesetFile in testSubscriptionB.filesOfStatus("Available"):
            assert filesetFile["lfn"] in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile["lfn"])

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        return

    def testAddBulkByLFN(self):
        """
        _testAddBulkByLFN_

        Verify that the Fileset.BulkAddByLFN DAO correct adds files to multiple
        filesets.
        """
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileA.addRun(Run(1, *[45]))
        testFileA.create()
        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileB.addRun(Run(1, *[45]))
        testFileB.create()
        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileC.addRun(Run(1, *[45]))
        testFileC.create()

        testFileD = File(lfn="/this/is/a/lfnD", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileD.addRun(Run(1, *[45]))
        testFileD.create()

        testWorkflowA = Workflow(spec="spec1.xml", owner="Hassen",
                                 name="wf001", task="sometask")
        testWorkflowA.create()

        testFilesetA = Fileset(name="TestFilesetA")
        testFilesetA.create()
        testFilesetB = Fileset(name="TestFilesetB")
        testFilesetB.create()

        testSubscriptionA = Subscription(fileset=testFilesetA,
                                         workflow=testWorkflowA)
        testSubscriptionA.create()
        testSubscriptionB = Subscription(fileset=testFilesetB,
                                         workflow=testWorkflowA)
        testSubscriptionB.create()

        myThread = threading.currentThread()
        myThread.transaction.begin()

        binds = [{"lfn": testFileA["lfn"], "fileset": testFilesetA.id},
                 {"lfn": testFileB["lfn"], "fileset": testFilesetA.id},
                 {"lfn": testFileC["lfn"], "fileset": testFilesetA.id},
                 {"lfn": testFileB["lfn"], "fileset": testFilesetB.id},
                 {"lfn": testFileC["lfn"], "fileset": testFilesetB.id},
                 {"lfn": testFileD["lfn"], "fileset": testFilesetB.id}]

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        bulkAddAction = daoFactory(classname="Fileset.BulkAddByLFN")
        bulkAddAction.execute(binds=binds)

        testFilesetA.loadData()
        testFilesetB.loadData()

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetA.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testSubscriptionA.filesOfStatus("Available"):
            assert filesetFile["lfn"] in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile["lfn"])

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileB, testFileC, testFileD]
        for filesetFile in testFilesetB.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileB, testFileC, testFileD]
        for filesetFile in testSubscriptionB.filesOfStatus("Available"):
            assert filesetFile["lfn"] in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile["lfn"])

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        return

    def testAddToWMBSInBulk(self):
        """
        testAddToWMBSInBulk

        test create and add files to fileset in one go
        """
        testFileA = File(lfn="/this/is/a/lfnA", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileA.addRun(Run(1, *[45]))

        testFileB = File(lfn="/this/is/a/lfnB", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileB.addRun(Run(1, *[45]))

        testFileC = File(lfn="/this/is/a/lfnC", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileC.addRun(Run(1, *[45]))

        testFileD = File(lfn="/this/is/a/lfnD", size=1024,
                         events=20, checksums={'cksum': 3},
                         locations=set(["T2_CH_CERN"]))
        testFileD.addRun(Run(1, *[45]))

        testWorkflowA = Workflow(spec="spec1.xml", owner="Hassen",
                                 name="wf001", task="sometask")
        testWorkflowA.create()
        testWorkflowB = Workflow(spec="spec2.xml", owner="Hassen",
                                 name="wf002", task="sometask")
        testWorkflowB.create()

        testFilesetA = Fileset(name="TestFilesetA")
        testFilesetA.create()
        testFilesetB = Fileset(name="TestFilesetB")
        testFilesetB.create()

        testSubscriptionA = Subscription(fileset=testFilesetA,
                                         workflow=testWorkflowA)
        testSubscriptionA.create()
        testSubscriptionB = Subscription(fileset=testFilesetB,
                                         workflow=testWorkflowB)
        testSubscriptionB.create()

        myThread = threading.currentThread()
        myThread.transaction.begin()

        testFilesetA.addFilesToWMBSInBulk([testFileA, testFileB, testFileC],
                                          testWorkflowA.name)
        testFilesetB.addFilesToWMBSInBulk([testFileB, testFileC, testFileD],
                                          testWorkflowB.name)

        testFilesetA.loadData()
        testFilesetB.loadData()

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testFilesetA.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileA, testFileB, testFileC]
        for filesetFile in testSubscriptionA.filesOfStatus("Available"):
            assert filesetFile["lfn"] in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile["lfn"])

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileB, testFileC, testFileD]
        for filesetFile in testFilesetB.files:
            assert filesetFile in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile)

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        goldenFiles = [testFileB, testFileC, testFileD]
        for filesetFile in testSubscriptionB.filesOfStatus("Available"):
            assert filesetFile["lfn"] in goldenFiles, \
                "ERROR: Unknown file in fileset"
            goldenFiles.remove(filesetFile["lfn"])

        assert len(goldenFiles) == 0, \
            "ERROR: Fileset is missing files"

        return

    def testSetLastUpdate(self):
        """
        _testSetLastUpdate_

        Test that setting the lastUpdate of a fileset works when changing it
        with the setLastUpdate() method.
        """
        testFilesetA = Fileset(name="TestFileset1")
        testFilesetA.create()
        testFilesetA.setLastUpdate(10)
        testFilesetB = Fileset(name="TestFileset2")
        testFilesetB.create()
        testFilesetB.setLastUpdate(20)

        testFilesetC = Fileset(name=testFilesetA.name)
        testFilesetC.load()
        testFilesetD = Fileset(name=testFilesetB.name)
        testFilesetD.load()

        assert testFilesetC.lastUpdate == 10, \
            "ERROR: lastUpdate of FilesetC should be 10."

        assert testFilesetD.lastUpdate == 20, \
            "ERROR: lastUpdate of FilesetD should be 20."

    def testListFilesetByTask(self):
        """
        _testListFilesetByTask_

        Verify that Fileset.ListFilesetByTask DAO correct turns
        the list of fileset by task.
        """
        testWorkflow1 = Workflow(spec="spec1.xml", owner="Hassen",
                                 name="wf001", task="sometask")
        testWorkflow1.create()
        testFilesetA = Fileset(name="TestFileset1")
        testFilesetA.create()
        testSubscription1 = Subscription(fileset=testFilesetA,
                                         workflow=testWorkflow1)
        testSubscription1.create()

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        listFilesetByTaskDAO = daoFactory(classname="Fileset.ListFilesetByTask")
        listFilesetByTask = listFilesetByTaskDAO.execute(task=testWorkflow1.task)

        assert len(listFilesetByTask) == 1, \
            "ERROR: listFilesetByTask should be 1."

        assert listFilesetByTask[0]['name'] == "TestFileset1", \
            "ERROR: the fileset  should be TestFileset1."

        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_FileBased_t_

File based splitting test.
"""




from builtins import range
import unittest

from WMCore.DataStructs.File import File
from WMCore.DataStructs.Fileset import Fileset
from WMCore.DataStructs.Job import Job
from WMCore.DataStructs.Subscription import Subscription
from WMCore.DataStructs.Workflow import Workflow
from WMCore.DataStructs.Run import Run

from WMCore.JobSplitting.SplitterFactory import SplitterFactory
from WMCore.Services.UUIDLib import makeUUID

class FileBasedTest(unittest.TestCase):
    """
    _FileBasedTest_

    Test file based job splitting.
    """


    def setUp(self):
        """
        _setUp_

        Create two subscriptions: One that contains a single file and one that
        contains multiple files.
        """
        self.multipleFileFileset = Fileset(name = "TestFileset1")
        for i in range(10):
            newFile = File(makeUUID(), size = 1000, events = 100)
            newFile.setLocation('blenheim')
            newFile.setLocation('malpaquet')
            lumis = []
            for lumi in range(20):
                lumis.append((i * 100) + lumi)
                newFile.addRun(Run(i, *lumis))
            self.multipleFileFileset.addFile(newFile)

        self.singleFileFileset = Fileset(name = "TestFileset2")
        newFile = File("/some/file/name", size = 1000, events = 100)
        newFile.setLocation('blenheim')
        lumis = list(range(50,60)) + list(range(70,80))
        newFile.addRun(Run(13, *lumis))
        self.singleFileFileset.addFile(newFile)

        testWorkflow = Workflow()
        self.multipleFileSubscription = Subscription(fileset = self.multipleFileFileset,
                                                     workflow = testWorkflow,
                                                     split_algo = "FileBased",
                                                     type = "Processing")
        self.singleFileSubscription = Subscription(fileset = self.singleFileFileset,
                                                   workflow = testWorkflow,
                                                   split_algo = "FileBased",
                                                   type = "Processing")

        #self.multipleFileSubscription.create()
        #self.singleFileSubscription.create()

        self.performanceParams = {'timePerEvent' : 12,
                                  'memoryRequirement' : 2300,
                                  'sizePerEvent' : 400}

        return

    def tearDown(self):
        """
        _tearDown_

        Nothing to do...
        """
        pass

    def testExactFiles(self):
        """
        _testExactFiles_

        Test file based job splitting when the number of files per job is
        exactly the same as the number of files in the input fileset.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)

        jobGroups = jobFactory(files_per_job = 1,
                               performance = self.performanceParams)

        assert len(jobGroups) == 1, \
               "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 1, \
               "ERROR: JobFactory didn't create a single job."

        job = jobGroups[0].jobs.pop()

        assert job.getFiles(type = "lfn") == ["/some/file/name"], \
               "ERROR: Job contains unknown files."

        return

    def testMoreFiles(self):
        """
        _testMoreFiles_

        Test file based job splitting when the number of files per job is
        greater than the number of files in the input fileset.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.singleFileSubscription)

        jobGroups = jobFactory(files_per_job = 10,
                               performance = self.performanceParams)

        assert len(jobGroups) == 1, \
               "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 1, \
               "ERROR: JobFactory didn't create a single job."

        job = jobGroups[0].jobs.pop()

        assert job.getFiles(type = "lfn") == ["/some/file/name"], \
               "ERROR: Job contains unknown files."

        return

    def test2FileSplit(self):
        """
        _test2FileSplit_

        Test file based job splitting when the number of files per job is
        2, this should result in five jobs.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.multipleFileSubscription)

        jobGroups = jobFactory(files_per_job = 2,
                               performance = self.performanceParams)

        assert len(jobGroups) == 1, \
               "ERROR: JobFactory didn't return one JobGroup."

        assert len(jobGroups[0].jobs) == 5, \
               "ERROR: JobFactory didn't create two jobs."

        fileSet = set()
        for job in jobGroups[0].jobs:
            assert len(job.getFiles(type = "set")) == 2, \
                   "ERROR: Job contains incorrect number of files."

            for file in job.getFiles(type = "lfn"):
                fileSet.add(file)

        assert len(fileSet) == 10, \
               "ERROR: Not all files assinged to job."

        return

    def test3FileSplit(self):
        """
        _test3FileSplit_

        Test file based job splitting when the number of files per job is
        3, this should result in four jobs.
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.multipleFileSubscription)

        jobGroups = jobFactory(files_per_job = 3,
                               performance = self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 4)

        fileList = []
        for job in jobGroups[0].jobs:
            assert len(job.getFiles(type = "list")) in [3, 1], \
                   "ERROR: Job contains incorrect number of files."

            for file in job.getFiles(type = "lfn"):
                assert file not in fileList, \
                       "ERROR: File duplicated!"
                fileList.append(file)

        self.assertEqual(len(fileList), 10)

        return

    def test4WithLumiMask(self):
        """
        _test4WithLumiMask_

        Test file based job splitting when
        """
        splitter = SplitterFactory()
        jobFactory = splitter(self.multipleFileSubscription)

        jobGroups = jobFactory(files_per_job = 2,
                               total_files = 3,
                               runs = ['1', '2', '4', '5'],
                               lumis = ['100,130', '203,204,207,221', '401,405', '500, 520'],
                               performance = self.performanceParams)

        self.assertEqual(len(jobGroups), 1)

        self.assertEqual(len(jobGroups[0].jobs), 2)

        fileList = []
        for job in jobGroups[0].jobs:
            assert len(job.getFiles(type = "list")) in [2, 1], \
                   "ERROR: Job contains incorrect number of files."

            for file in job.getFiles(type = "lfn"):
                assert file not in fileList, \
                       "ERROR: File duplicated!"
                fileList.append(file)

        self.assertEqual(len(fileList), 3)

        return

if __name__ == '__main__':
    unittest.main()



from __future__ import absolute_import
from __future__ import print_function
import unittest

from .WorkQueueTestCase import WorkQueueTestCase

class WorkQueueSchemnaTest(WorkQueueTestCase):
    """
    _WorkQueueSchemnaTest_

    test schema creation and deletion
    """

    def testSchemaGeneration(self):
        print("test schema creation")

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
    WorkQueue.Policy.Start.MonteCarlo tests
"""
from __future__ import division

import math
import unittest

from WMCore_t.WMSpec_t.samples.MultiMergeProductionWorkload import workload as MultiMergeProductionWorkload
from WMCore_t.WMSpec_t.samples.MultiTaskProductionWorkload import workload as MultiTaskProductionWorkload
from WMCore_t.WorkQueue_t.WorkQueue_t import getFirstTask

from WMCore.WorkQueue.Policy.Start.MonteCarlo import MonteCarlo
from WMCore.WorkQueue.WorkQueueExceptions import WorkQueueNoWorkError, WorkQueueWMSpecError
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.Emulators.WMSpecGenerator.Samples.BasicProductionWorkload import getProdArgs, taskChainWorkload


mcArgs = getProdArgs()


class MonteCarloTestCase(EmulatedUnitTestCase):
    """Test case MonteCarlo Workload"""

    splitArgs = dict(SliceType='NumEvents', SliceSize=100, MaxJobsPerElement=1)

    def testBasicProductionWorkload(self):
        """Basic Production Workload"""
        # change split defaults for this test
        totalevents = 1000000
        splitArgs = dict(SliceType='NumberOfEvents', SliceSize=100, MaxJobsPerElement=5)
        BasicProductionWorkload = taskChainWorkload('MonteCarloWorkload', mcArgs)
        getFirstTask(BasicProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        getFirstTask(BasicProductionWorkload).addProduction(totalEvents=totalevents)
        getFirstTask(BasicProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        for task in BasicProductionWorkload.taskIterator():
            units, _, _ = MonteCarlo(**splitArgs)(BasicProductionWorkload, task)

            SliceSize = BasicProductionWorkload.startPolicyParameters()['SliceSize']
            self.assertEqual(math.ceil(totalevents / (SliceSize * splitArgs['MaxJobsPerElement'])), len(units))
            first_event = 1
            first_lumi = 1
            first_run = 1
            for unit in units:
                self.assertTrue(unit['Jobs'] <= splitArgs['MaxJobsPerElement'])
                self.assertEqual(unit['WMSpec'], BasicProductionWorkload)
                self.assertEqual(unit['Task'], task)
                self.assertEqual(unit['Mask']['FirstEvent'], first_event)
                self.assertEqual(unit['Mask']['FirstLumi'], first_lumi)
                last_event = first_event + (SliceSize * unit['Jobs']) - 1
                if last_event > totalevents:
                    # this should be the last unit of work
                    last_event = totalevents
                last_lumi = first_lumi + unit['Jobs'] - 1
                self.assertEqual(unit['Mask']['LastEvent'], last_event)
                self.assertEqual(unit['Mask']['LastLumi'], last_lumi)
                self.assertEqual(unit['Mask']['FirstRun'], first_run)
                self.assertEqual(last_lumi - first_lumi + 1, unit['NumberOfLumis'])
                self.assertEqual(last_event - first_event + 1, unit['NumberOfEvents'])
                first_event = last_event + 1
                first_lumi += unit['Jobs']  # one lumi per job
            self.assertEqual(unit['Mask']['LastEvent'], totalevents)

    def testLHEProductionWorkload(self):
        """
        _testLHEProductionWorkload_

        Make sure that splitting by event plus events in per lumi works

        """
        totalevents = 542674
        splitArgs = dict(SliceType='NumEvents', SliceSize=47, MaxJobsPerElement=5, SubSliceType='NumEventsPerLumi',
                         SubSliceSize=13)

        LHEProductionWorkload = taskChainWorkload('MonteCarloWorkload', mcArgs)
        LHEProductionWorkload.setJobSplittingParameters(
            getFirstTask(LHEProductionWorkload).getPathName(),
            'EventBased',
            {'events_per_job': splitArgs['SliceSize'],
             'events_per_lumi': splitArgs['SubSliceSize']})
        getFirstTask(LHEProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        getFirstTask(LHEProductionWorkload).addProduction(totalEvents=totalevents)
        getFirstTask(LHEProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        for task in LHEProductionWorkload.taskIterator():
            units, _, _ = MonteCarlo(**splitArgs)(LHEProductionWorkload, task)

            SliceSize = LHEProductionWorkload.startPolicyParameters()['SliceSize']
            self.assertEqual(math.ceil(totalevents / (SliceSize * splitArgs['MaxJobsPerElement'])), len(units))
            first_event = 1
            first_lumi = 1
            first_run = 1
            lumis_per_job = int(math.ceil(SliceSize / splitArgs['SubSliceSize']))
            for unit in units:
                self.assertTrue(unit['Jobs'] <= splitArgs['MaxJobsPerElement'])
                self.assertEqual(unit['WMSpec'], LHEProductionWorkload)
                self.assertEqual(unit['Task'], task)
                self.assertEqual(unit['Mask']['FirstEvent'], first_event)
                self.assertEqual(unit['Mask']['FirstLumi'], first_lumi)
                last_event = first_event + (SliceSize * unit['Jobs']) - 1
                last_lumi = first_lumi + (lumis_per_job * unit['Jobs']) - 1
                if last_event > totalevents:
                    # this should be the last unit of work
                    last_event = totalevents
                    last_lumi = first_lumi
                    last_lumi += math.ceil(((last_event - first_event + 1) % SliceSize) / splitArgs['SubSliceSize'])
                    last_lumi += (lumis_per_job * (unit['Jobs'] - 1)) - 1
                self.assertEqual(unit['Mask']['LastEvent'], last_event)
                self.assertEqual(unit['Mask']['LastLumi'], last_lumi)
                self.assertEqual(unit['Mask']['FirstRun'], first_run)
                first_event = last_event + 1
                first_lumi = last_lumi + 1
            self.assertEqual(unit['Mask']['LastEvent'], totalevents)

    def testExtremeSplits(self):
        """
        _testExtremeSplits_

        Make sure that the protection to avoid going over 2^32 works

        """
        totalevents = 2 ** 34
        splitArgs = dict(SliceType='NumEvents', SliceSize=2 ** 30, MaxJobsPerElement=7)

        LHEProductionWorkload = taskChainWorkload('MonteCarloWorkload', mcArgs)
        LHEProductionWorkload.setJobSplittingParameters(
            getFirstTask(LHEProductionWorkload).getPathName(),
            'EventBased',
            {'events_per_job': splitArgs['SliceSize'],
             'events_per_lumi': splitArgs['SliceSize']})
        getFirstTask(LHEProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        getFirstTask(LHEProductionWorkload).addProduction(totalEvents=totalevents)
        getFirstTask(LHEProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        for task in LHEProductionWorkload.taskIterator():
            units, _, _ = MonteCarlo(**splitArgs)(LHEProductionWorkload, task)

            SliceSize = LHEProductionWorkload.startPolicyParameters()['SliceSize']
            self.assertEqual(math.ceil(totalevents / (SliceSize * splitArgs['MaxJobsPerElement'])), len(units))
            self.assertEqual(len(units), 3, "Should produce 3 units")

            unit1 = units[0]
            unit2 = units[1]
            unit3 = units[2]

            self.assertEqual(unit1['Jobs'], 7, 'First unit produced more jobs than expected')
            self.assertEqual(unit1['Mask']['FirstEvent'], 1, 'First unit has a wrong first event')
            self.assertEqual(unit1['Mask']['LastEvent'], 7 * (2 ** 30), 'First unit has a wrong last event')
            self.assertEqual(unit2['Jobs'], 7, 'Second unit produced more jobs than expected')
            self.assertEqual(unit2['Mask']['FirstEvent'], 2 ** 30 + 1, 'Second unit has a wrong first event')
            self.assertEqual(unit2['Mask']['LastEvent'], 8 * (2 ** 30), 'Second unit has a wrong last event')
            self.assertEqual(unit3['Jobs'], 2, 'Third unit produced more jobs than expected')
            self.assertEqual(unit3['Mask']['FirstEvent'], 2 * (2 ** 30) + 1, 'Third unit has a wrong first event')
            self.assertEqual(unit3['Mask']['LastEvent'], 4 * (2 ** 30), 'First unit has a wrong last event')

    def testShiftedStartSplitting(self):
        """
        _testShiftedStartSplitting_

        Make sure that splitting by event plus events in per lumi works
        when the first event and lumi is not 1

        """
        totalevents = 542674
        splitArgs = dict(SliceType='NumEvents', SliceSize=47, MaxJobsPerElement=5, SubSliceType='NumEventsPerLumi',
                         SubSliceSize=13)

        LHEProductionWorkload = taskChainWorkload('MonteCarloWorkload', mcArgs)
        LHEProductionWorkload.setJobSplittingParameters(getFirstTask(LHEProductionWorkload).getPathName(), 'EventBased',
                                                        {'events_per_job': splitArgs['SliceSize'],
                                                         'events_per_lumi': splitArgs['SubSliceSize']})
        getFirstTask(LHEProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        getFirstTask(LHEProductionWorkload).addProduction(totalEvents=totalevents)
        getFirstTask(LHEProductionWorkload).setFirstEventAndLumi(50, 100)
        getFirstTask(LHEProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        for task in LHEProductionWorkload.taskIterator():
            units, _, _ = MonteCarlo(**splitArgs)(LHEProductionWorkload, task)

            SliceSize = LHEProductionWorkload.startPolicyParameters()['SliceSize']
            self.assertEqual(math.ceil(totalevents / (SliceSize * splitArgs['MaxJobsPerElement'])), len(units))
            first_event = 50
            first_lumi = 100
            first_run = 1
            lumis_per_job = int(math.ceil(SliceSize / splitArgs['SubSliceSize']))
            for unit in units:
                self.assertTrue(unit['Jobs'] <= splitArgs['MaxJobsPerElement'])
                self.assertEqual(unit['WMSpec'], LHEProductionWorkload)
                self.assertEqual(unit['Task'], task)
                self.assertEqual(unit['Mask']['FirstEvent'], first_event)
                self.assertEqual(unit['Mask']['FirstLumi'], first_lumi)
                last_event = first_event + (SliceSize * unit['Jobs']) - 1
                last_lumi = first_lumi + (lumis_per_job * unit['Jobs']) - 1
                if last_event > totalevents:
                    # this should be the last unit of work
                    last_event = totalevents + 50 - 1
                    last_lumi = first_lumi
                    last_lumi += math.ceil(((last_event - first_event + 1) % SliceSize) / splitArgs['SubSliceSize'])
                    last_lumi += (lumis_per_job * (unit['Jobs'] - 1)) - 1
                self.assertEqual(unit['Mask']['LastEvent'], last_event)
                self.assertEqual(unit['Mask']['LastLumi'], last_lumi)
                self.assertEqual(unit['Mask']['FirstRun'], first_run)
                first_event = last_event + 1
                first_lumi = last_lumi + 1
            self.assertEqual(unit['Mask']['LastEvent'], totalevents + 50 - 1)

    def testMultiMergeProductionWorkload(self):
        """Multi merge production workload"""
        getFirstTask(MultiMergeProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        getFirstTask(MultiMergeProductionWorkload).setFirstEventAndLumi(1, 1)
        for task in MultiMergeProductionWorkload.taskIterator():
            units, _, _ = MonteCarlo(**self.splitArgs)(MultiMergeProductionWorkload, task)

            self.assertEqual(10.0, len(units))
            for unit in units:
                self.assertEqual(1, unit['Jobs'])
                self.assertEqual(unit['WMSpec'], MultiMergeProductionWorkload)
                self.assertEqual(unit['Task'], task)

    def testMultiTaskProcessingWorkload(self):
        """Multi Task Processing Workflow"""
        count = 0
        getFirstTask(MultiTaskProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])
        for task in MultiTaskProductionWorkload.taskIterator():
            count += 1
            task.setFirstEventAndLumi(1, 1)
            units, _, _ = MonteCarlo(**self.splitArgs)(MultiTaskProductionWorkload, task)

            self.assertEqual(10 * count, len(units))
            for unit in units:
                self.assertEqual(1, unit['Jobs'])
                self.assertEqual(unit['WMSpec'], MultiTaskProductionWorkload)
                self.assertEqual(unit['Task'], task)
        self.assertEqual(count, 2)

    def testInvalidSpecs(self):
        """Specs with no work"""
        mcspec = taskChainWorkload('testProcessingInvalid', mcArgs)
        mcspec.setSiteWhitelist(["T2_XX_SiteA"])
        # 0 events
        getFirstTask(mcspec).addProduction(totalEvents=0)
        for task in mcspec.taskIterator():
            self.assertRaises(WorkQueueNoWorkError, MonteCarlo(), mcspec, task)

        # -ve split size
        mcspec2 = taskChainWorkload('testProdInvalid', mcArgs)
        mcspec2.data.policies.start.SliceSize = -100
        for task in mcspec2.taskIterator():
            self.assertRaises(WorkQueueWMSpecError, MonteCarlo(), mcspec2, task)

    def testContinuousSplittingSupport(self):
        """Doesn't support continuous splitting"""
        policyInstance = MonteCarlo(**self.splitArgs)
        self.assertFalse(policyInstance.supportsWorkAddition(),
                         "MonteCarlo instance should not support continuous splitting")
        self.assertRaises(NotImplementedError, policyInstance.newDataAvailable, *[None, None])
        self.assertRaises(NotImplementedError, policyInstance.modifyPolicyForWorkAddition, *[None])
        return

    def _getLHEProductionWorkload(self, splitArgs):

        totalevents = 1010

        LHEProductionWorkload = taskChainWorkload('MonteCarloWorkload', mcArgs)
        LHEProductionWorkload.setJobSplittingParameters(
            getFirstTask(LHEProductionWorkload).getPathName(),
            'EventBased',
            {'events_per_job': splitArgs['SliceSize'],
             'events_per_lumi': splitArgs['SubSliceSize']})
        getFirstTask(LHEProductionWorkload).addProduction(totalEvents=totalevents)
        getFirstTask(LHEProductionWorkload).setSiteWhitelist(['T2_XX_SiteA', 'T2_XX_SiteB'])

        return LHEProductionWorkload

    def testJobsPerElementProductionWorkload(self):
        """test dynaminc changes on MaxJobPerElements"""

        splitArgs = dict(SliceType='NumEvents', SliceSize=100, MaxJobsPerElement=8, SubSliceType='NumEventsPerLumi',
                         SubSliceSize=11, MaxLumisPerElement=100)

        LHEProductionWorkload = self._getLHEProductionWorkload(splitArgs)
        for task in LHEProductionWorkload.taskIterator():
            mc = MonteCarlo(**splitArgs)
            mc(LHEProductionWorkload, task)
            self.assertEqual(mc.args["MaxJobsPerElement"], 8)

        splitArgs = dict(SliceType='NumEvents', SliceSize=100, MaxJobsPerElement=8, SubSliceType='NumEventsPerLumi',
                         SubSliceSize=11, MaxLumisPerElement=40)

        LHEProductionWorkload = self._getLHEProductionWorkload(splitArgs)
        for task in LHEProductionWorkload.taskIterator():
            mc = MonteCarlo(**splitArgs)
            mc(LHEProductionWorkload, task)
            self.assertEqual(mc.args["MaxJobsPerElement"], 4)

        splitArgs = dict(SliceType='NumEvents', SliceSize=100, MaxJobsPerElement=8, SubSliceType='NumEventsPerLumi',
                         SubSliceSize=11, MaxLumisPerElement=1)

        LHEProductionWorkload = self._getLHEProductionWorkload(splitArgs)
        for task in LHEProductionWorkload.taskIterator():
            mc = MonteCarlo(**splitArgs)
            with self.assertRaises(WorkQueueWMSpecError):
                mc(LHEProductionWorkload, task)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_AgentStatusWatcher_

This component is a general purpose component for resources and processes handling
"""

import logging
import threading

from WMCore.Agent.Harness import Harness
from WMComponent.AgentStatusWatcher.ResourceControlUpdater import ResourceControlUpdater
from WMComponent.AgentStatusWatcher.AgentStatusPoller import AgentStatusPoller
from WMComponent.AgentStatusWatcher.DrainStatusPoller import DrainStatusPoller

class AgentStatusWatcher(Harness):
    """
    Component class for AgentStatusWatcher module
    """

    def __init__(self, config):
        """
        __init__

        Initialize the Harness
        """
        Harness.__init__(self, config)

        return

    def preInitialization(self):
        """
        _preInitialization_

        Sets up the worker thread
        """
        logging.info("AgentStatusWatcher.preInitialization")
        resourceUpdaterPollInterval = self.config.AgentStatusWatcher.resourceUpdaterPollInterval
        agentPollInterval = self.config.AgentStatusWatcher.agentPollInterval
        drainStatusPollInterval = self.config.AgentStatusWatcher.drainStatusPollInterval
        myThread = threading.currentThread()

        logging.info("Setting AgentStatusPoller poll interval to %s seconds", agentPollInterval)
        myThread.workerThreadManager.addWorker(AgentStatusPoller(self.config),
                                               agentPollInterval)

        logging.info("Setting ResourceControlUpdater poll interval to %s seconds", resourceUpdaterPollInterval)
        myThread.workerThreadManager.addWorker(ResourceControlUpdater(self.config),
                                               resourceUpdaterPollInterval)

        if not hasattr(self.config, "Tier0Feeder"):
            # Don't set up DrainStausPoller for Tier0
            logging.info("Setting DrainStatusPoller poll interval to %s seconds", drainStatusPollInterval)
            myThread.workerThreadManager.addWorker(DrainStatusPoller(self.config),
                                               drainStatusPollInterval)
        return



#!/usr/bin/env python
"""
_BasicProductionWorkload_

Sample/Tester for a simple production workflow and associated merge.

Production task produces a single output dataset,
Merge task is used to merge that dataset

"""
from WMCore.WMSpec.WMWorkload import newWorkload
from WMCore.WMSpec.StdSpecs.TaskChain import TaskChainWorkloadFactory

#  //
# // Set up the basic workload task and step structure
#//
def createWorkload(name="BasicProduction"):
    workload = newWorkload(name)
    workload.setOwner("DMWMTest")
    workload.setStartPolicy('MonteCarlo')
    workload.setEndPolicy('SingleShot')

    #  //
    # // set up the production task
    #//
    production = workload.newTask("Production")
    #TODO: Currently WMBS only support Processing, Merge, Harvesting type - may be add production type?
    production.setTaskType("Merge")

    production.addProduction(totalevents = 1000)
    prodCmssw = production.makeStep("cmsRun1")
    prodCmssw.setStepType("CMSSW")
    prodStageOut = prodCmssw.addStep("stageOut1")
    prodStageOut.setStepType("StageOut")
    production.applyTemplates()
    production.setSiteWhitelist(["T2_XX_SiteA"])

    #  //
    # // set up the merge task
    #//
    merge = production.addTask("Merge")
    mergeCmssw = merge.makeStep("cmsRun1")
    mergeCmssw.setStepType("CMSSW")
    mergeStageOut = mergeCmssw.addStep("stageOut1")
    mergeStageOut.setStepType("StageOut")
    merge.applyTemplates()


    #  //
    # // populate the details of the production tasks
    #//
    #  //
    # // production cmssw step
    #//
    #
    # TODO: Anywhere helper.data is accessed means we need a method added to the
    # type based helper class to provide a clear API.
    prodCmsswHelper = prodCmssw.getTypeHelper()

    prodCmsswHelper.data.application.setup.cmsswVersion = "CMSSW_X_Y_Z"
    prodCmsswHelper.data.application.setup.softwareEnvironment = " . /uscmst1/prod/sw/cms/bashrc prod"

    prodCmsswHelper.data.application.configuration.configCacheUrl = "http://whatever"

    prodCmsswHelper.addOutputModule("writeData", primaryDataset = "Primary",
                                    processedDataset = "Processed",
                                    dataTier = "TIER")
    #print prodCmsswHelper.data



    #  //
    # // production stage out step
    #//
    prodStageOutHelper = prodStageOut.getTypeHelper()


    #  //
    # // merge cmssw step
    #//
    # point it at the input step from the previous task
    merge.setInputReference(prodCmssw, outputModule = "writeData")

    return workload


def getProdArgs():
    mcArgs = TaskChainWorkloadFactory.getTestArguments()
    mcArgs.update({
                   "CouchURL": None,
                   "CouchDBName": None,
                   "ConfigCacheDoc" : None
                   })
    mcArgs.pop('ConfigCacheDoc')

    return mcArgs


class TestTaskChainFactory(TaskChainWorkloadFactory):
    """Override bits that talk to cmsssw"""
    def __call__(self, workflowName, args):
        workload = TaskChainWorkloadFactory.__call__(self, workflowName, args)
        #delattr(workload.taskIterator().next().steps().data.application.configuration,
        #        'configCacheUrl')
        return workload

    def determineOutputModules(self, *args, **kwargs):
        "Don't talk to couch"
        return {}

def taskChainWorkload(workloadName, arguments):
    """
    _monteCarloWorkload_
    Instantiate the MonteCarloWorkflowFactory and have it generate a workload for
    the given parameters.
    """
    myTaskChainFactory = TestTaskChainFactory()
    return myTaskChainFactory(workloadName, arguments)


#!/usr/bin/env python
"""
_LogCollect_

Builder implementation for LogCollect

"""
from __future__ import print_function

import logging

from WMCore.WMSpec.ConfigSectionTree import nodeName
from WMCore.WMSpec.Steps.Builder import Builder


class LogCollect(Builder):
    """
    _LogCollect_

    Build a working area for a LogCollect step

    """

    def build(self, step, workingDir, **args):
        """
        _build_

        implement build interface for LogCollect Step

        """
        stepName = nodeName(step)
        stepWorkingArea = "%s/%s" % (workingDir, stepName)
        self.installWorkingArea(step, stepWorkingArea)
        logging.info("Builders.LogCollect.build called on %s", stepName)



#!/usr/bin/env python

"""
JobArchiver test
"""

from builtins import range
from future.utils import listvalues

import os
import shutil
import threading
import unittest
from subprocess import PIPE, Popen

from nose.plugins.attrib import attr

from WMComponent.JobArchiver.JobArchiverPoller import JobArchiverPoller
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.Services.UUIDLib import makeUUID
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMQuality.Emulators import EmulatorSetup
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit


class JobArchiverTest(EmulatedUnitTestCase):
    """
    TestCase for TestJobArchiver module
    """

    _maxMessage = 10

    def setUp(self):
        """
        setup for test.
        """
        super(JobArchiverTest, self).setUp()

        myThread = threading.currentThread()
        super(JobArchiverTest, self).setUp()
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        # self.tearDown()
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)
        self.testInit.setupCouch("jobarchiver_t_0/jobs", "JobDump")
        self.testInit.setupCouch("jobarchiver_t_0/fwjrs", "FWJRDump")

        self.daofactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.getJobs = self.daofactory(classname="Jobs.GetAllJobs")

        self.testDir = self.testInit.generateWorkDir(deleteOnDestruction=False)

        self.nJobs = 10

        self.configFile = EmulatorSetup.setupWMAgentConfig()

        return

    def tearDown(self):
        """
        Database deletion
        """
        self.testInit.clearDatabase(modules=["WMCore.WMBS"])
        self.testInit.tearDownCouch()
        self.testInit.delWorkDir()
        EmulatorSetup.deleteConfig(self.configFile)
        super(JobArchiverTest, self).tearDown()

        return

    def getConfig(self):
        """
        _createConfig_

        General config file
        """
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        # First the general stuff
        config.section_("General")
        config.General.workDir = os.getenv("TESTDIR", os.getcwd())
        config.General.WorkDir = os.getenv("TESTDIR", os.getcwd())

        # Now the CoreDatabase information
        # This should be the dialect, dburl, etc
        config.section_("CoreDatabase")
        config.CoreDatabase.connectUrl = os.getenv("DATABASE")
        config.CoreDatabase.socket = os.getenv("DBSOCK")

        config.section_("JobStateMachine")
        config.JobStateMachine.couchurl = os.getenv("COUCHURL", "cmssrv48.fnal.gov:5984")
        config.JobStateMachine.couchDBName = "jobarchiver_t_0"

        config.component_("JobArchiver")
        config.JobArchiver.pollInterval = 60
        config.JobArchiver.logLevel = 'INFO'
        # config.JobArchiver.logDir                = os.path.join(self.testDir, 'logs')
        config.JobArchiver.componentDir = self.testDir
        config.JobArchiver.numberOfJobsToCluster = 1000

        config.component_('WorkQueueManager')
        config.WorkQueueManager.namespace = "WMComponent.WorkQueueManager.WorkQueueManager"
        config.WorkQueueManager.componentDir = config.General.workDir + "/WorkQueueManager"
        config.WorkQueueManager.level = 'LocalQueue'
        config.WorkQueueManager.logLevel = 'DEBUG'
        config.WorkQueueManager.couchurl = 'https://None'
        config.WorkQueueManager.dbname = 'whatever'
        config.WorkQueueManager.inboxDatabase = 'whatever2'
        config.WorkQueueManager.queueParams = {}
        config.WorkQueueManager.queueParams["ParentQueueCouchUrl"] = "https://cmsweb.cern.ch/couchdb/workqueue"

        return config

    def createTestJobGroup(self):
        """
        Creates a group of several jobs

        """

        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileA.addRun(Run(10, *[12312]))
        testFileA.setLocation('malpaquet')

        testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        testFileB.addRun(Run(10, *[12312]))
        testFileA.setLocation('malpaquet')
        testFileA.create()
        testFileB.create()

        testWMBSFileset.addFile(testFileA)
        testWMBSFileset.addFile(testFileB)
        testWMBSFileset.commit()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        for _ in range(0, self.nJobs):
            testJob = Job(name=makeUUID())
            testJob.addFile(testFileA)
            testJob.addFile(testFileB)
            testJob['retry_count'] = 1
            testJob['retry_max'] = 10
            testJobGroup.add(testJob)

        testJobGroup.commit()

        return testJobGroup

    def testBasicFunctionTest(self):
        """
        _BasicFunctionTest_

        Tests the components, by seeing if they can process a simple set of closeouts
        """

        myThread = threading.currentThread()

        config = self.getConfig()

        testJobGroup = self.createTestJobGroup()

        changer = ChangeState(config)

        cacheDir = os.path.join(self.testDir, 'test')

        if not os.path.isdir(cacheDir):
            os.mkdir(cacheDir)

        # if os.path.isdir(config.JobArchiver.logDir):
        #    shutil.rmtree(config.JobArchiver.logDir)

        for job in testJobGroup.jobs:
            myThread.transaction.begin()
            job["outcome"] = "success"
            job.save()
            myThread.transaction.commit()
            path = os.path.join(cacheDir, job['name'])
            os.makedirs(path)
            f = open('%s/%s.out' % (path, job['name']), 'w')
            f.write(job['name'])
            f.close()
            job.setCache(path)

        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')
        changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        changer.propagate(testJobGroup.jobs, 'success', 'complete')

        testJobArchiver = JobArchiverPoller(config=config)
        testJobArchiver.algorithm()

        result = myThread.dbi.processData(
            "SELECT wmbs_job_state.name FROM wmbs_job_state INNER JOIN wmbs_job ON wmbs_job.state = wmbs_job_state.id")[
            0].fetchall()

        for val in result:
            self.assertEqual(listvalues(val), ['cleanout'])

        dirList = os.listdir(cacheDir)
        for job in testJobGroup.jobs:
            self.assertEqual(job["name"] in dirList, False)

        logPath = os.path.join(config.JobArchiver.componentDir, 'logDir', 'w', 'wf001', 'JobCluster_0')
        logList = os.listdir(logPath)
        for job in testJobGroup.jobs:
            self.assertEqual('Job_%i.tar.bz2' % (job['id']) in logList, True,
                             'Could not find transferred tarball for job %i' % (job['id']))
            pipe = Popen(['tar', '-jxvf', os.path.join(logPath, 'Job_%i.tar.bz2' % (job['id']))],
                         stdout=PIPE, stderr=PIPE, shell=False)
            pipe.wait()
            # filename = '%s/%s/%s.out' %(cacheDir[1:], job['name'], job['name'])
            filename = 'Job_%i/%s.out' % (job['id'], job['name'])
            self.assertEqual(os.path.isfile(filename), True, 'Could not find file %s' % (filename))
            f = open(filename, 'r')
            fileContents = f.readlines()
            f.close()
            self.assertEqual(fileContents[0].find(job['name']) > -1, True)
            shutil.rmtree('Job_%i' % (job['id']))
            if os.path.isfile('Job_%i.tar.bz2' % (job['id'])):
                os.remove('Job_%i.tar.bz2' % (job['id']))

        return

    @attr('integration')
    def testSpeedTest(self):
        """
        _SpeedTest_

        Tests the components, as in sees if they load.
        Otherwise does nothing.
        """
        return
        # What used to be here - stop pylint from complaining about unreachable code
        # myThread = threading.currentThread()
        #
        # config = self.getConfig()
        #
        # self.nJobs = 2000
        #
        # testJobGroup = self.createTestJobGroup()
        #
        # changer = ChangeState(config)
        #
        # cacheDir = os.path.join(self.testDir, 'test')
        #
        # for job in testJobGroup.jobs:
        #     job["outcome"] = "success"
        #     job.save()
        #     path = os.path.join(cacheDir, job['name'])
        #     os.makedirs(path)
        #     f = open('%s/%s.out' % (path, job['name']), 'w')
        #     f.write(job['name'])
        #     f.close()
        #     job.setCache(path)
        #
        # changer.propagate(testJobGroup.jobs, 'created', 'new')
        # changer.propagate(testJobGroup.jobs, 'executing', 'created')
        # changer.propagate(testJobGroup.jobs, 'complete', 'executing')
        # changer.propagate(testJobGroup.jobs, 'success', 'complete')
        #
        # testJobArchiver = JobArchiverPoller(config=config)
        # cProfile.runctx("testJobArchiver.algorithm()", globals(), locals(), filename="testStats.stat")
        #
        # p = pstats.Stats('testStats.stat')
        # p.sort_stats('cumulative')
        # p.print_stats(.2)
        #
        # return


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_ReReco_t_

Unit tests for the ReReco workflow.
"""
from __future__ import print_function
from future.utils import viewitems

import os
import threading
import unittest

from Utils.PythonVersion import PY3

from WMCore.DAOFactory import DAOFactory
from WMCore.Database.CMSCouch import CouchServer, Document
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMSpec.StdSpecs.ReReco import ReRecoWorkloadFactory
from WMCore.WorkQueue.WMBSHelper import WMBSHelper
from WMQuality.TestInitCouchApp import TestInitCouchApp



class ReRecoTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        Initialize the database and couch.
        """
        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setupCouch("rereco_t", "ConfigCache")
        self.testInit.setSchema(customModules=["WMCore.WMBS"],
                                useDefault=False)
        self.testDir = self.testInit.generateWorkDir()
        couchServer = CouchServer(os.environ["COUCHURL"])
        self.configDatabase = couchServer.connectDatabase("rereco_t")

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.listTasksByWorkflow = self.daoFactory(classname="Workflow.LoadFromName")
        self.listFilesets = self.daoFactory(classname="Fileset.List")
        self.listSubsMapping = self.daoFactory(classname="Subscriptions.ListSubsAndFilesetsFromWorkflow")
        if PY3:
            self.assertItemsEqual = self.assertCountEqual
        return

    def tearDown(self):
        """
        _tearDown_

        Clear out the database.
        """
        self.testInit.tearDownCouch()
        self.testInit.clearDatabase()
        self.testInit.delWorkDir()
        return

    def injectReRecoConfig(self):
        """
        _injectReRecoConfig_

        Inject a ReReco config document that we can use to set the outputModules
        """

        newConfig = Document()
        newConfig["info"] = None
        newConfig["config"] = None
        newConfig["md5hash"] = "eb1c38cf50e14cf9fc31278a5c8e580f"
        newConfig["pset_hash"] = "7c856ad35f9f544839d8525ca10259a7"
        newConfig["owner"] = {"group": "cmsdataops", "user": "sfoulkes"}
        newConfig["pset_tweak_details"] = {"process": {"outputModules_": ['RECOoutput', 'DQMoutput'],
                                                       "RECOoutput": {'dataset': {'filterName': 'RECOoutputFilter',
                                                                                  'dataTier': 'RECO'}},
                                                       "DQMoutput": {'dataset': {'filterName': 'DQMoutputFilter',
                                                                                 'dataTier': 'DQM'}}}}
        result = self.configDatabase.commitOne(newConfig)
        return result[0]["id"]

    def injectDQMHarvestConfig(self):
        """
        _injectDQMHarvest_

        Create a bogus config cache document for DQMHarvest and
        inject it into couch.  Return the ID of the document.
        """
        newConfig = Document()
        newConfig["info"] = None
        newConfig["config"] = None
        newConfig["md5hash"] = "eb1c38cf50e14cf9fc31278a5c8e234f"
        newConfig["pset_hash"] = "7c856ad35f9f544839d8525ca10876a7"
        newConfig["owner"] = {"group": "DATAOPS", "user": "amaltaro"}
        newConfig["pset_tweak_details"] = {"process": {"outputModules_": []}}
        result = self.configDatabase.commitOne(newConfig)
        return result[0]["id"]

    def injectSkimConfig(self):
        """
        _injectSkimConfig_

        Create a bogus config cache document for the skims and inject it into
        couch.  Return the ID of the document.
        """
        newConfig = Document()
        newConfig["info"] = None
        newConfig["config"] = None
        newConfig["md5hash"] = "eb1c38cf50e14cf9fc31278a5c8e580f"
        newConfig["pset_hash"] = "7c856ad35f9f544839d8525ca10259a7"
        newConfig["owner"] = {"group": "cmsdataops", "user": "sfoulkes"}
        newConfig["pset_tweak_details"] = {"process": {"outputModules_": ["SkimA", "SkimB"],
                                                       "SkimA": {"dataset": {"filterName": "SkimAFilter",
                                                                             "dataTier": "RAW-RECO"}},
                                                       "SkimB": {"dataset": {"filterName": "SkimBFilter",
                                                                             "dataTier": "USER"}}}}
        result = self.configDatabase.commitOne(newConfig)
        return result[0]["id"]

    def testReReco(self):
        """
        _testReReco_

        Verify that ReReco workflows can be created and inserted into WMBS
        correctly.
        """
        skimConfig = self.injectSkimConfig()
        recoConfig = self.injectReRecoConfig()
        dataProcArguments = ReRecoWorkloadFactory.getTestArguments()
        dataProcArguments["ProcessingString"] = "ProcString"
        dataProcArguments["ConfigCacheID"] = recoConfig
        dataProcArguments.update({"SkimName1": "SomeSkim",
                                  "SkimInput1": "RECOoutput",
                                  "Skim1ConfigCacheID": skimConfig})
        dataProcArguments["CouchURL"] = os.environ["COUCHURL"]
        dataProcArguments["CouchDBName"] = "rereco_t"
        dataProcArguments["EnableHarvesting"] = True
        dataProcArguments["DQMConfigCacheID"] = self.injectDQMHarvestConfig()

        factory = ReRecoWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", dataProcArguments)

        self.assertEqual(testWorkload.data.tasks.DataProcessing.tree.children.DataProcessingMergeRECOoutput. \
                         tree.children.SomeSkim.tree.children.SomeSkimMergeSkimB.steps.cmsRun1.output.modules. \
                         Merged.mergedLFNBase,
                         '/store/data/FAKE/MinimumBias/USER/SkimBFilter-ProcString-v1')

        testWMBSHelper = WMBSHelper(testWorkload, "DataProcessing", "SomeBlock", cachepath=self.testDir)
        testWMBSHelper.createTopLevelFileset()
        testWMBSHelper._createSubscriptionsInWMBS(testWMBSHelper.topLevelTask, testWMBSHelper.topLevelFileset)

        procWorkflow = Workflow(name="TestWorkload",
                                task="/TestWorkload/DataProcessing")
        procWorkflow.load()

        self.assertEqual(len(procWorkflow.outputMap), 3,
                         "Error: Wrong number of WF outputs.")

        goldenOutputMods = {"RECOoutput": "RECO", "DQMoutput": "DQM"}
        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            fset = goldenOutputMod + tier
            mergedOutput = procWorkflow.outputMap[fset][0]["merged_output_fileset"]
            unmergedOutput = procWorkflow.outputMap[fset][0]["output_fileset"]
            mergedOutput.loadData()
            unmergedOutput.loadData()

            self.assertEqual(mergedOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Merged output fileset is wrong: %s" % mergedOutput.name)
            self.assertEqual(unmergedOutput.name, "/TestWorkload/DataProcessing/unmerged-%s" % fset,
                             "Error: Unmerged output fileset is wrong.")

        logArchOutput = procWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
        unmergedLogArchOutput = procWorkflow.outputMap["logArchive"][0]["output_fileset"]
        logArchOutput.loadData()
        unmergedLogArchOutput.loadData()

        self.assertEqual(logArchOutput.name, "/TestWorkload/DataProcessing/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")
        self.assertEqual(unmergedLogArchOutput.name, "/TestWorkload/DataProcessing/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            mergeWorkflow = Workflow(name="TestWorkload",
                                     task="/TestWorkload/DataProcessing/DataProcessingMerge%s" % goldenOutputMod)
            mergeWorkflow.load()

            self.assertEqual(len(mergeWorkflow.outputMap), 2,
                             "Error: Wrong number of WF outputs.")

            mergedMergeOutput = mergeWorkflow.outputMap["Merged%s" % tier][0]["merged_output_fileset"]
            unmergedMergeOutput = mergeWorkflow.outputMap["Merged%s" % tier][0]["output_fileset"]

            mergedMergeOutput.loadData()
            unmergedMergeOutput.loadData()

            self.assertEqual(mergedMergeOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Merged output fileset is wrong.")
            self.assertEqual(unmergedMergeOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Unmerged output fileset is wrong.")

            logArchOutput = mergeWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
            unmergedLogArchOutput = mergeWorkflow.outputMap["logArchive"][0]["output_fileset"]
            logArchOutput.loadData()
            unmergedLogArchOutput.loadData()

            self.assertEqual(logArchOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMerge%s/merged-logArchive" % goldenOutputMod,
                             "Error: LogArchive output fileset is wrong: %s" % logArchOutput.name)
            self.assertEqual(unmergedLogArchOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMerge%s/merged-logArchive" % goldenOutputMod,
                             "Error: LogArchive output fileset is wrong.")

        topLevelFileset = Fileset(name="TestWorkload-DataProcessing-SomeBlock")
        topLevelFileset.loadData()

        procSubscription = Subscription(fileset=topLevelFileset, workflow=procWorkflow)
        procSubscription.loadData()

        self.assertEqual(procSubscription["type"], "Processing",
                         "Error: Wrong subscription type.")
        self.assertEqual(procSubscription["split_algo"], "EventAwareLumiBased",
                         "Error: Wrong split algo.")

        unmergedReco = Fileset(name="/TestWorkload/DataProcessing/unmerged-RECOoutputRECO")
        unmergedReco.loadData()
        recoMergeWorkflow = Workflow(name="TestWorkload",
                                     task="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput")
        recoMergeWorkflow.load()
        mergeSubscription = Subscription(fileset=unmergedReco, workflow=recoMergeWorkflow)
        mergeSubscription.loadData()

        self.assertEqual(mergeSubscription["type"], "Merge",
                         "Error: Wrong subscription type.")
        self.assertEqual(mergeSubscription["split_algo"], "ParentlessMergeBySize",
                         "Error: Wrong split algo.")

        unmergedDqm = Fileset(name="/TestWorkload/DataProcessing/unmerged-DQMoutputDQM")
        unmergedDqm.loadData()
        dqmMergeWorkflow = Workflow(name="TestWorkload",
                                    task="/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput")
        dqmMergeWorkflow.load()
        mergeSubscription = Subscription(fileset=unmergedDqm, workflow=dqmMergeWorkflow)
        mergeSubscription.loadData()

        self.assertEqual(mergeSubscription["type"], "Merge",
                         "Error: Wrong subscription type.")
        self.assertEqual(mergeSubscription["split_algo"], "ParentlessMergeBySize",
                         "Error: Wrong split algo.")

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            fset = goldenOutputMod + tier
            unmerged = Fileset(name="/TestWorkload/DataProcessing/unmerged-%s" % fset)
            unmerged.loadData()
            cleanupWorkflow = Workflow(name="TestWorkload",
                                       task="/TestWorkload/DataProcessing/DataProcessingCleanupUnmerged%s" % goldenOutputMod)
            cleanupWorkflow.load()
            cleanupSubscription = Subscription(fileset=unmerged, workflow=cleanupWorkflow)
            cleanupSubscription.loadData()

            self.assertEqual(cleanupSubscription["type"], "Cleanup",
                             "Error: Wrong subscription type.")
            self.assertEqual(cleanupSubscription["split_algo"], "SiblingProcessingBased",
                             "Error: Wrong split algo.")

        procLogCollect = Fileset(name="/TestWorkload/DataProcessing/unmerged-logArchive")
        procLogCollect.loadData()
        procLogCollectWorkflow = Workflow(name="TestWorkload",
                                          task="/TestWorkload/DataProcessing/LogCollect")
        procLogCollectWorkflow.load()
        logCollectSub = Subscription(fileset=procLogCollect, workflow=procLogCollectWorkflow)
        logCollectSub.loadData()

        self.assertEqual(logCollectSub["type"], "LogCollect",
                         "Error: Wrong subscription type.")
        self.assertEqual(logCollectSub["split_algo"], "MinFileBased",
                         "Error: Wrong split algo.")

        procLogCollect = Fileset(name="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/merged-logArchive")
        procLogCollect.loadData()
        procLogCollectWorkflow = Workflow(name="TestWorkload",
                                          task="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/DataProcessingRECOoutputMergeLogCollect")
        procLogCollectWorkflow.load()
        logCollectSub = Subscription(fileset=procLogCollect, workflow=procLogCollectWorkflow)
        logCollectSub.loadData()

        self.assertEqual(logCollectSub["type"], "LogCollect",
                         "Error: Wrong subscription type.")
        self.assertEqual(logCollectSub["split_algo"], "MinFileBased",
                         "Error: Wrong split algo.")

        procLogCollect = Fileset(name="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/merged-logArchive")
        procLogCollect.loadData()
        procLogCollectWorkflow = Workflow(name="TestWorkload",
                                          task="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/DataProcessingRECOoutputMergeLogCollect")
        procLogCollectWorkflow.load()
        logCollectSub = Subscription(fileset=procLogCollect, workflow=procLogCollectWorkflow)
        logCollectSub.loadData()

        self.assertEqual(logCollectSub["type"], "LogCollect",
                         "Error: Wrong subscription type.")
        self.assertEqual(logCollectSub["split_algo"], "MinFileBased",
                         "Error: Wrong split algo.")

        skimWorkflow = Workflow(name="TestWorkload",
                                task="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim")
        skimWorkflow.load()

        self.assertEqual(len(skimWorkflow.outputMap), 3,
                         "Error: Wrong number of WF outputs.")

        goldenOutputMods = {"SkimA": "RAW-RECO", "SkimB": "USER"}
        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            fset = goldenOutputMod + tier
            mergedOutput = skimWorkflow.outputMap[fset][0]["merged_output_fileset"]
            unmergedOutput = skimWorkflow.outputMap[fset][0]["output_fileset"]

            mergedOutput.loadData()
            unmergedOutput.loadData()

            self.assertEqual(mergedOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Merged output fileset is wrong: %s" % mergedOutput.name)
            self.assertEqual(unmergedOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-%s" % fset,
                             "Error: Unmerged output fileset is wrong: %s" % unmergedOutput.name)

        logArchOutput = skimWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
        unmergedLogArchOutput = skimWorkflow.outputMap["logArchive"][0]["output_fileset"]
        logArchOutput.loadData()
        unmergedLogArchOutput.loadData()

        self.assertEqual(logArchOutput.name,
                         "/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")
        self.assertEqual(unmergedLogArchOutput.name,
                         "/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            mergeWorkflow = Workflow(name="TestWorkload",
                                     task="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s" % goldenOutputMod)
            mergeWorkflow.load()

            self.assertEqual(len(mergeWorkflow.outputMap), 2,
                             "Error: Wrong number of WF outputs.")

            mergedMergeOutput = mergeWorkflow.outputMap["Merged%s" % tier][0]["merged_output_fileset"]
            unmergedMergeOutput = mergeWorkflow.outputMap["Merged%s" % tier][0]["output_fileset"]

            mergedMergeOutput.loadData()
            unmergedMergeOutput.loadData()

            self.assertEqual(mergedMergeOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Merged output fileset is wrong.")
            self.assertEqual(unmergedMergeOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Unmerged output fileset is wrong.")

            logArchOutput = mergeWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
            unmergedLogArchOutput = mergeWorkflow.outputMap["logArchive"][0]["output_fileset"]
            logArchOutput.loadData()
            unmergedLogArchOutput.loadData()

            self.assertEqual(logArchOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s/merged-logArchive" % goldenOutputMod,
                             "Error: LogArchive output fileset is wrong: %s" % logArchOutput.name)
            self.assertEqual(unmergedLogArchOutput.name,
                             "/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s/merged-logArchive" % goldenOutputMod,
                             "Error: LogArchive output fileset is wrong.")

        topLevelFileset = Fileset(name="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/merged-MergedRECO")
        topLevelFileset.loadData()

        skimSubscription = Subscription(fileset=topLevelFileset, workflow=skimWorkflow)
        skimSubscription.loadData()

        self.assertEqual(skimSubscription["type"], "Skim",
                         "Error: Wrong subscription type.")
        self.assertEqual(skimSubscription["split_algo"], "FileBased",
                         "Error: Wrong split algo.")

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            fset = goldenOutputMod + tier
            unmerged = Fileset(
                name="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-%s" % fset)
            unmerged.loadData()
            mergeWorkflow = Workflow(name="TestWorkload",
                                     task="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s" % goldenOutputMod)
            mergeWorkflow.load()
            mergeSubscription = Subscription(fileset=unmerged, workflow=mergeWorkflow)
            mergeSubscription.loadData()

            self.assertEqual(mergeSubscription["type"], "Merge",
                             "Error: Wrong subscription type.")
            self.assertEqual(mergeSubscription["split_algo"], "ParentlessMergeBySize",
                             "Error: Wrong split algo.")

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            fset = goldenOutputMod + tier
            unmerged = Fileset(
                name="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-%s" % fset)
            unmerged.loadData()
            cleanupWorkflow = Workflow(name="TestWorkload",
                                       task="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimCleanupUnmerged%s" % goldenOutputMod)
            cleanupWorkflow.load()
            cleanupSubscription = Subscription(fileset=unmerged, workflow=cleanupWorkflow)
            cleanupSubscription.loadData()

            self.assertEqual(cleanupSubscription["type"], "Cleanup",
                             "Error: Wrong subscription type.")
            self.assertEqual(cleanupSubscription["split_algo"], "SiblingProcessingBased",
                             "Error: Wrong split algo.")

        for skimOutput in goldenOutputMods:
            skimMergeLogCollect = Fileset(
                name="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s/merged-logArchive" % skimOutput)
            skimMergeLogCollect.loadData()
            skimMergeLogCollectWorkflow = Workflow(name="TestWorkload",
                                                   task="/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMerge%s/SomeSkim%sMergeLogCollect" % (
                                                       skimOutput, skimOutput))
            skimMergeLogCollectWorkflow.load()
            logCollectSub = Subscription(fileset=skimMergeLogCollect, workflow=skimMergeLogCollectWorkflow)
            logCollectSub.loadData()

            self.assertEqual(logCollectSub["type"], "LogCollect",
                             "Error: Wrong subscription type.")
            self.assertEqual(logCollectSub["split_algo"], "MinFileBased",
                             "Error: Wrong split algo.")

        dqmWorkflow = Workflow(name="TestWorkload",
                               task="/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged")
        dqmWorkflow.load()

        topLevelFileset = Fileset(name="/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/merged-MergedDQM")
        topLevelFileset.loadData()

        dqmSubscription = Subscription(fileset=topLevelFileset, workflow=dqmWorkflow)
        dqmSubscription.loadData()

        self.assertEqual(dqmSubscription["type"], "Harvesting",
                         "Error: Wrong subscription type.")
        self.assertEqual(dqmSubscription["split_algo"], "Harvest",
                         "Error: Wrong split algo.")

        logArchOutput = dqmWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
        unmergedLogArchOutput = dqmWorkflow.outputMap["logArchive"][0]["output_fileset"]
        logArchOutput.loadData()
        unmergedLogArchOutput.loadData()

        self.assertEqual(logArchOutput.name,
                         "/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")
        self.assertEqual(unmergedLogArchOutput.name,
                         "/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")

        dqmHarvestLogCollect = Fileset(
            name="/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged/unmerged-logArchive")
        dqmHarvestLogCollect.loadData()
        dqmHarvestLogCollectWorkflow = Workflow(name="TestWorkload",
                                                task="/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged/DataProcessingMergeDQMoutputMergedEndOfRunDQMHarvestLogCollect")
        dqmHarvestLogCollectWorkflow.load()

        logCollectSub = Subscription(fileset=dqmHarvestLogCollect, workflow=dqmHarvestLogCollectWorkflow)
        logCollectSub.loadData()

        self.assertEqual(logCollectSub["type"], "LogCollect",
                         "Error: Wrong subscription type.")
        self.assertEqual(logCollectSub["split_algo"], "MinFileBased",
                         "Error: Wrong split algo.")

        return

    def testReRecoDroppingRECO(self):
        """
        _testReRecoDroppingRECO_

        Verify that ReReco workflows can be created and inserted into WMBS
        correctly.  The ReReco workflow is just a DataProcessing workflow with
        skims tacked on. This tests run on unmerged RECO output
        """
        skimConfig = self.injectSkimConfig()
        recoConfig = self.injectReRecoConfig()
        dataProcArguments = ReRecoWorkloadFactory.getTestArguments()
        dataProcArguments["ProcessingString"] = "ProcString"
        dataProcArguments["ConfigCacheID"] = recoConfig
        dataProcArguments.update({"SkimName1": "SomeSkim",
                                  "SkimInput1": "RECOoutput",
                                  "Skim1ConfigCacheID": skimConfig})
        dataProcArguments["CouchURL"] = os.environ["COUCHURL"]
        dataProcArguments["CouchDBName"] = "rereco_t"
        dataProcArguments["TransientOutputModules"] = ["RECOoutput"]
        dataProcArguments["EnableHarvesting"] = True
        dataProcArguments["DQMConfigCacheID"] = self.injectDQMHarvestConfig()

        factory = ReRecoWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", dataProcArguments)

        self.assertEqual(testWorkload.data.tasks.DataProcessing.tree.children. \
                         SomeSkim.tree.children.SomeSkimMergeSkimB.steps.cmsRun1.output.modules. \
                         Merged.mergedLFNBase,
                         '/store/data/FAKE/MinimumBias/USER/SkimBFilter-ProcString-v1')

        testWMBSHelper = WMBSHelper(testWorkload, "DataProcessing", "SomeBlock", cachepath=self.testDir)
        testWMBSHelper.createTopLevelFileset()
        testWMBSHelper._createSubscriptionsInWMBS(testWMBSHelper.topLevelTask, testWMBSHelper.topLevelFileset)

        skimWorkflow = Workflow(name="TestWorkload",
                                task="/TestWorkload/DataProcessing/SomeSkim")
        skimWorkflow.load()

        self.assertEqual(len(skimWorkflow.outputMap), 3,
                         "Error: Wrong number of WF outputs.")

        goldenOutputMods = {"SkimA": "RAW-RECO", "SkimB": "USER"}
        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            fset = goldenOutputMod + tier
            mergedOutput = skimWorkflow.outputMap[fset][0]["merged_output_fileset"]
            unmergedOutput = skimWorkflow.outputMap[fset][0]["output_fileset"]

            mergedOutput.loadData()
            unmergedOutput.loadData()

            self.assertEqual(mergedOutput.name,
                             "/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Merged output fileset is wrong: %s" % mergedOutput.name)
            self.assertEqual(unmergedOutput.name, "/TestWorkload/DataProcessing/SomeSkim/unmerged-%s" % fset,
                             "Error: Unmerged output fileset is wrong: %s" % unmergedOutput.name)

        logArchOutput = skimWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
        unmergedLogArchOutput = skimWorkflow.outputMap["logArchive"][0]["output_fileset"]
        logArchOutput.loadData()
        unmergedLogArchOutput.loadData()

        self.assertEqual(logArchOutput.name, "/TestWorkload/DataProcessing/SomeSkim/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")
        self.assertEqual(unmergedLogArchOutput.name, "/TestWorkload/DataProcessing/SomeSkim/unmerged-logArchive",
                         "Error: LogArchive output fileset is wrong.")

        for goldenOutputMod, tier in viewitems(goldenOutputMods):
            mergeWorkflow = Workflow(name="TestWorkload",
                                     task="/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s" % goldenOutputMod)
            mergeWorkflow.load()

            self.assertEqual(len(mergeWorkflow.outputMap), 2,
                             "Error: Wrong number of WF outputs.")

            mergedMergeOutput = mergeWorkflow.outputMap["Merged%s" % tier][0]["merged_output_fileset"]
            unmergedMergeOutput = mergeWorkflow.outputMap["Merged%s" % tier][0]["output_fileset"]

            mergedMergeOutput.loadData()
            unmergedMergeOutput.loadData()

            self.assertEqual(mergedMergeOutput.name,
                             "/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Merged output fileset is wrong.")
            self.assertEqual(unmergedMergeOutput.name,
                             "/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s/merged-Merged%s" % (
                             goldenOutputMod, tier),
                             "Error: Unmerged output fileset is wrong.")

            logArchOutput = mergeWorkflow.outputMap["logArchive"][0]["merged_output_fileset"]
            unmergedLogArchOutput = mergeWorkflow.outputMap["logArchive"][0]["output_fileset"]
            logArchOutput.loadData()
            unmergedLogArchOutput.loadData()

            self.assertEqual(logArchOutput.name,
                             "/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s/merged-logArchive" % goldenOutputMod,
                             "Error: LogArchive output fileset is wrong: %s" % logArchOutput.name)
            self.assertEqual(unmergedLogArchOutput.name,
                             "/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s/merged-logArchive" % goldenOutputMod,
                             "Error: LogArchive output fileset is wrong.")

        topLevelFileset = Fileset(name="/TestWorkload/DataProcessing/unmerged-RECOoutputRECO")
        topLevelFileset.loadData()

        skimSubscription = Subscription(fileset=topLevelFileset, workflow=skimWorkflow)
        skimSubscription.loadData()

        self.assertEqual(skimSubscription["type"], "Skim",
                         "Error: Wrong subscription type.")
        self.assertEqual(skimSubscription["split_algo"], "FileBased",
                         "Error: Wrong split algo.")

        for skimOutput, tier in viewitems(goldenOutputMods):
            fset = skimOutput + tier
            unmerged = Fileset(name="/TestWorkload/DataProcessing/SomeSkim/unmerged-%s" % fset)
            unmerged.loadData()
            mergeWorkflow = Workflow(name="TestWorkload",
                                     task="/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s" % skimOutput)
            mergeWorkflow.load()
            mergeSubscription = Subscription(fileset=unmerged, workflow=mergeWorkflow)
            mergeSubscription.loadData()

            self.assertEqual(mergeSubscription["type"], "Merge",
                             "Error: Wrong subscription type.")
            self.assertEqual(mergeSubscription["split_algo"], "ParentlessMergeBySize",
                             "Error: Wrong split algo.")

        for skimOutput, tier in viewitems(goldenOutputMods):
            fset = skimOutput + tier
            unmerged = Fileset(name="/TestWorkload/DataProcessing/SomeSkim/unmerged-%s" % fset)
            unmerged.loadData()
            cleanupWorkflow = Workflow(name="TestWorkload",
                                       task="/TestWorkload/DataProcessing/SomeSkim/SomeSkimCleanupUnmerged%s" % skimOutput)
            cleanupWorkflow.load()
            cleanupSubscription = Subscription(fileset=unmerged, workflow=cleanupWorkflow)
            cleanupSubscription.loadData()

            self.assertEqual(cleanupSubscription["type"], "Cleanup",
                             "Error: Wrong subscription type.")
            self.assertEqual(cleanupSubscription["split_algo"], "SiblingProcessingBased",
                             "Error: Wrong split algo.")

        for skimOutput in goldenOutputMods:
            skimMergeLogCollect = Fileset(
                name="/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s/merged-logArchive" % skimOutput)
            skimMergeLogCollect.loadData()
            skimMergeLogCollectWorkflow = Workflow(name="TestWorkload",
                                                   task="/TestWorkload/DataProcessing/SomeSkim/SomeSkimMerge%s/SomeSkim%sMergeLogCollect" % (
                                                       skimOutput, skimOutput))
            skimMergeLogCollectWorkflow.load()
            logCollectSub = Subscription(fileset=skimMergeLogCollect, workflow=skimMergeLogCollectWorkflow)
            logCollectSub.loadData()

            self.assertEqual(logCollectSub["type"], "LogCollect",
                             "Error: Wrong subscription type.")
            self.assertEqual(logCollectSub["split_algo"], "MinFileBased",
                             "Error: Wrong split algo.")

        return

    def testMemCoresSettings(self):
        """
        _testMemCoresSettings_

        Make sure the multicore and memory setings are properly propagated to
        all tasks and steps.
        """
        skimConfig = self.injectSkimConfig()
        recoConfig = self.injectReRecoConfig()
        dataProcArguments = ReRecoWorkloadFactory.getTestArguments()
        dataProcArguments["ConfigCacheID"] = recoConfig
        dataProcArguments.update({"SkimName1": "SomeSkim",
                                  "SkimInput1": "RECOoutput",
                                  "Skim1ConfigCacheID": skimConfig})
        dataProcArguments["CouchURL"] = os.environ["COUCHURL"]
        dataProcArguments["CouchDBName"] = "rereco_t"
        dataProcArguments["EnableHarvesting"] = True
        dataProcArguments["DQMConfigCacheID"] = self.injectDQMHarvestConfig()

        factory = ReRecoWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", dataProcArguments)

        # test default values
        taskPaths = {'/TestWorkload/DataProcessing': ['cmsRun1', 'stageOut1', 'logArch1'],
                     '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim': ['cmsRun1', 'stageOut1',
                                                                                             'logArch1'],
                     '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged': [
                         'cmsRun1', 'upload1', 'logArch1']}
        for task in taskPaths:
            taskObj = testWorkload.getTaskByPath(task)
            for step in taskPaths[task]:
                stepHelper = taskObj.getStepHelper(step)
                self.assertEqual(stepHelper.getNumberOfCores(), 1)
                self.assertEqual(stepHelper.getNumberOfStreams(), 0)
            # FIXME: not sure whether we should set performance parameters to Harvest jobs?!?
            if task == '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged':
                continue
            # then test Memory requirements
            perfParams = taskObj.jobSplittingParameters()['performance']
            self.assertEqual(perfParams['memoryRequirement'], 2300.0)

        # now test case where args are provided
        dataProcArguments["Multicore"] = 6
        dataProcArguments["Memory"] = 4600.0
        dataProcArguments["EventStreams"] = 3
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", dataProcArguments)
        for task in taskPaths:
            taskObj = testWorkload.getTaskByPath(task)
            for step in taskPaths[task]:
                stepHelper = taskObj.getStepHelper(step)
                if not task.endswith('DQMHarvestMerged') and step == 'cmsRun1':
                    self.assertEqual(stepHelper.getNumberOfCores(), dataProcArguments["Multicore"])
                    self.assertEqual(stepHelper.getNumberOfStreams(), dataProcArguments["EventStreams"])
                elif step in ('stageOut1', 'upload1', 'logArch1'):
                    self.assertEqual(stepHelper.getNumberOfCores(), 1)
                    self.assertEqual(stepHelper.getNumberOfStreams(), 0)
                else:
                    self.assertEqual(stepHelper.getNumberOfCores(), 1, "%s should be single-core" % task)
                    self.assertEqual(stepHelper.getNumberOfStreams(), 0)
            # FIXME: not sure whether we should set performance parameters to Harvest jobs?!?
            if task == '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged':
                continue
            # then test Memory requirements
            perfParams = taskObj.jobSplittingParameters()['performance']
            self.assertEqual(perfParams['memoryRequirement'], dataProcArguments["Memory"])

        return

    def testFilesets(self):
        """
        _testFilesets_

        Test workflow tasks, filesets and subscriptions creation
        """
        # expected tasks, filesets, subscriptions, etc
        expOutTasks = ['/TestWorkload/DataProcessing',
                       '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput',
                       '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim',
                       '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimA',
                       '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimB',
                       '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput']
        expWfTasks = ['/TestWorkload/DataProcessing',
                      '/TestWorkload/DataProcessing/DataProcessingCleanupUnmergedDQMoutput',
                      '/TestWorkload/DataProcessing/DataProcessingCleanupUnmergedRECOoutput',
                      '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput',
                      '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingDQMoutputMergeLogCollect',
                      '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged',
                      '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged/DataProcessingMergeDQMoutputMergedEndOfRunDQMHarvestLogCollect',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/DataProcessingRECOoutputMergeLogCollect',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimCleanupUnmergedSkimA',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimCleanupUnmergedSkimB',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimLogCollect',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimA',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimA/SomeSkimSkimAMergeLogCollect',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimB',
                      '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimB/SomeSkimSkimBMergeLogCollect',
                      '/TestWorkload/DataProcessing/LogCollect']
        expFsets = ['TestWorkload-DataProcessing-/MinimumBias/ComissioningHI-v1/RAW',
                    '/TestWorkload/DataProcessing/unmerged-RECOoutputRECO',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/merged-MergedRECO',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimA/merged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimA/merged-MergedRAW-RECO',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimB/merged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimB/merged-MergedUSER',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-SkimARAW-RECO',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-SkimBUSER',
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged/unmerged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/merged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/merged-MergedDQM',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/merged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-logArchive',
                    '/TestWorkload/DataProcessing/unmerged-DQMoutputDQM',
                    '/TestWorkload/DataProcessing/unmerged-logArchive']
        subMaps = [(5,
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged/unmerged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged/DataProcessingMergeDQMoutputMergedEndOfRunDQMHarvestLogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (6,
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/merged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingDQMoutputMergeLogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (4,
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/merged-MergedDQM',
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput/DataProcessingMergeDQMoutputEndOfRunDQMHarvestMerged',
                    'Harvest',
                    'Harvesting'),
                   (17,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/merged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/DataProcessingRECOoutputMergeLogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (9,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/merged-MergedRECO',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim',
                    'FileBased',
                    'Skim'),
                   (12,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimA/merged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimA/SomeSkimSkimAMergeLogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (15,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimB/merged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimB/SomeSkimSkimBMergeLogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (16,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-logArchive',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimLogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (10,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-SkimARAW-RECO',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimCleanupUnmergedSkimA',
                    'SiblingProcessingBased',
                    'Cleanup'),
                   (11,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-SkimARAW-RECO',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimA',
                    'ParentlessMergeBySize',
                    'Merge'),
                   (13,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-SkimBUSER',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimCleanupUnmergedSkimB',
                    'SiblingProcessingBased',
                    'Cleanup'),
                   (14,
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/unmerged-SkimBUSER',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput/SomeSkim/SomeSkimMergeSkimB',
                    'ParentlessMergeBySize',
                    'Merge'),
                   (2,
                    '/TestWorkload/DataProcessing/unmerged-DQMoutputDQM',
                    '/TestWorkload/DataProcessing/DataProcessingCleanupUnmergedDQMoutput',
                    'SiblingProcessingBased',
                    'Cleanup'),
                   (3,
                    '/TestWorkload/DataProcessing/unmerged-DQMoutputDQM',
                    '/TestWorkload/DataProcessing/DataProcessingMergeDQMoutput',
                    'ParentlessMergeBySize',
                    'Merge'),
                   (18,
                    '/TestWorkload/DataProcessing/unmerged-logArchive',
                    '/TestWorkload/DataProcessing/LogCollect',
                    'MinFileBased',
                    'LogCollect'),
                   (7,
                    '/TestWorkload/DataProcessing/unmerged-RECOoutputRECO',
                    '/TestWorkload/DataProcessing/DataProcessingCleanupUnmergedRECOoutput',
                    'SiblingProcessingBased',
                    'Cleanup'),
                   (8,
                    '/TestWorkload/DataProcessing/unmerged-RECOoutputRECO',
                    '/TestWorkload/DataProcessing/DataProcessingMergeRECOoutput',
                    'ParentlessMergeBySize',
                    'Merge'),
                   (1,
                    'TestWorkload-DataProcessing-/MinimumBias/ComissioningHI-v1/RAW',
                    '/TestWorkload/DataProcessing',
                    'EventAwareLumiBased',
                    'Processing')]

        testArguments = ReRecoWorkloadFactory.getTestArguments()
        testArguments["ConfigCacheID"] = self.injectReRecoConfig()
        testArguments.update({"SkimName1": "SomeSkim",
                              "SkimInput1": "RECOoutput",
                              "Skim1ConfigCacheID": self.injectSkimConfig()})
        testArguments["CouchURL"] = os.environ["COUCHURL"]
        testArguments["CouchDBName"] = "rereco_t"
        testArguments["EnableHarvesting"] = True
        testArguments["DQMConfigCacheID"] = self.injectDQMHarvestConfig()

        factory = ReRecoWorkloadFactory()
        testWorkload = factory.factoryWorkloadConstruction("TestWorkload", testArguments)

        testWMBSHelper = WMBSHelper(testWorkload, "DataProcessing", blockName=testArguments['InputDataset'],
                                    cachepath=self.testInit.testDir)
        testWMBSHelper.createTopLevelFileset()
        testWMBSHelper._createSubscriptionsInWMBS(testWMBSHelper.topLevelTask, testWMBSHelper.topLevelFileset)

        self.assertItemsEqual(testWorkload.listOutputProducingTasks(), expOutTasks)

        workflows = self.listTasksByWorkflow.execute(workflow="TestWorkload")
        self.assertItemsEqual([item['task'] for item in workflows], expWfTasks)

        # returns a tuple of id, name, open and last_update
        filesets = self.listFilesets.execute()
        self.assertItemsEqual([item[1] for item in filesets], expFsets)

        subscriptions = self.listSubsMapping.execute(workflow="TestWorkload", returnTuple=True)
        self.assertItemsEqual(subscriptions, subMaps)


if __name__ == '__main__':
    unittest.main()
